# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Any
from ..classes.search import ConjunctionSearch
from ._swarmaurora import create_custom_import_file as func_create_custom_import_file
from ._swarmaurora import get_url as func_get_url
from ._swarmaurora import open_in_browser as func_open_in_browser

__all__ = ["SwarmAuroraManager"]


class SwarmAuroraManager:
    """
    The SwarmAuroraManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

    def get_url(self, search_obj: ConjunctionSearch) -> str:
        """
        Get a URL that displays a conjunction search in the Swarm-Aurora
        Conjunction Finder

        Args:
            search_obj (ConjunctionSearch): 
                A conjunction search object, must be a completed search with 
                the 'request_id' value populated

        Returns:
            The Swarm-Aurora Conjunction Finder URL for this conjunction search
        """
        return func_get_url(search_obj)

    def open_in_browser(self, search_obj: ConjunctionSearch, browser: Optional[str] = None) -> None:
        """
        In a browser, open a conjunction search in the Swarm-Aurora Conjunction Finder.

        Args:
            search_obj (ConjunctionSearch): 
                A conjunction search object, must be a completed search with 
                the 'request_id' value populated

            browser (str): 
                The browser type to load using. Default is your default browser. Some common other 
                options are "google-chrome", "firefox", or "safari". For all available options, refer 
                to https://docs.python.org/3/library/webbrowser.html#webbrowser.get
        
        Raises:
            ValueError: Unsupported browser specified
        """
        return func_open_in_browser(search_obj, browser)

    def create_custom_import_file(self, search_obj: ConjunctionSearch, filename: Optional[str] = None, return_dict: bool = False) -> Any:
        """
        Generate a Swarm-Aurora custom import file for a given conjunction search

        Args:
            search_obj (ConjunctionSearch): 
                A conjunction search object, must be a completed search with 
                the 'request_id' value populated

            filename (str): 
                The output filename, default is 'swarmaurora_custom_import_file_{requestID}.json'
            
            return_dict (bool): 
                Return the custom import file contents as a dictionary instead of saving 
                a file, default is False

        Returns:
            The filename of the saved custom import file, or a dictionary with the file contents 
            if `return_dict` is set to True
        """
        return func_create_custom_import_file(self.__aurorax_obj, search_obj, filename, return_dict)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class representation for a mosaic.
"""

import os
import datetime
import pyproj
import aacgmv2
import matplotlib.colors
import numpy as np
import matplotlib.pyplot as plt
import cartopy.feature
import cartopy.crs
from copy import copy
from dataclasses import dataclass
from typing import List, Dict, Tuple, Sequence, Union, Optional, Any
from numpy import ndarray
from matplotlib.collections import PolyCollection
from cartopy.crs import Projection
from ..._util import show_warning


@dataclass
class MosaicSkymap:
    """
    Prepared skymap data for use by mosaic routines.

    Attributes:
        site_uid_list (List[str]): 
            List of site unique identifiers contained within this object.

        elevation (List[numpy.ndarray]): 
            List of elevation data, with each element corresponding to each site. Order 
            matches that of the `site_uid_list` attribute.

        polyfill_lat (List[numpy.ndarray]): 
            List of latitude polygon data, with each element corresponding to each site. 
            Order matches that of the `site_uid_list` attribute. 

        polyfill_lon (List[numpy.ndarray]): 
            List of longitude polygon data, with each element corresponding to each site. 
            Order matches that of the `site_uid_list` attribute. 

    """

    site_uid_list: List[str]
    elevation: List[ndarray]
    polyfill_lat: List[ndarray]
    polyfill_lon: List[ndarray]

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # set special strings
        unique_polyfill_lat_dims = str(list(dict.fromkeys(fill_arr.shape
                                                          for fill_arr in self.polyfill_lat))).replace("[", "").replace("]",
                                                                                                                        "").replace("), (", "),(")
        unique_polyfill_lon_dims = str(list(dict.fromkeys(fill_arr.shape
                                                          for fill_arr in self.polyfill_lon))).replace("[", "").replace("]",
                                                                                                                        "").replace("), (", "),(")
        unique_elevation_dims = str(list(dict.fromkeys(el.shape for el in self.elevation))).replace("[", "").replace("]", "").replace("), (", "),(")

        polyfill_lat_str = "array(dims=%s, dtype=%s)" % (unique_polyfill_lat_dims, self.polyfill_lat[0].dtype)
        polyfill_lon_str = "array(dims=%s, dtype=%s)" % (unique_polyfill_lon_dims, self.polyfill_lon[0].dtype)
        elevation_str = "array(dims=%s, dtype=%s)" % (unique_elevation_dims, self.elevation[0].dtype)

        # return
        return "MosaicSkymap(polyfill_lat=%s, polyfill_lon=%s, elevation=%s, site_uid_list=%s)" % (
            polyfill_lat_str,
            polyfill_lon_str,
            elevation_str,
            self.site_uid_list.__repr__(),
        )

    def pretty_print(self):
        """
        A special print output for this class.
        """
        unique_polyfill_lat_dims = str(list(dict.fromkeys(fill_arr.shape
                                                          for fill_arr in self.polyfill_lat))).replace("[", "").replace("]",
                                                                                                                        "").replace("), (", "),(")
        unique_polyfill_lon_dims = str(list(dict.fromkeys(fill_arr.shape
                                                          for fill_arr in self.polyfill_lon))).replace("[", "").replace("]",
                                                                                                                        "").replace("), (", "),(")
        unique_elevation_dims = str(list(dict.fromkeys(el.shape for el in self.elevation))).replace("[", "").replace("]", "").replace("), (", "),(")

        polyfill_lat_str = "array(dims=%s, dtype=%s)" % (unique_polyfill_lat_dims, self.polyfill_lat[0].dtype)
        polyfill_lon_str = "array(dims=%s, dtype=%s)" % (unique_polyfill_lon_dims, self.polyfill_lon[0].dtype)
        elevation_str = "array(dims=%s, dtype=%s)" % (unique_elevation_dims, self.elevation[0].dtype)

        # print
        print("MosaicSkymap:")
        print("  %-15s: %s" % ("polyfill_lat", polyfill_lat_str))
        print("  %-15s: %s" % ("polyfill_lon", polyfill_lon_str))
        print("  %-15s: %s" % ("elevation", elevation_str))
        print("  %-15s: %s" % ("site_uid_list", self.site_uid_list))


@dataclass
class MosaicData:
    """
    Prepared image data for use by mosaic routines.

    Attributes:
        site_uid_list (List[str]): 
            List of site unique identifiers contained within this object.

        timestamps (List[datetime.datetime]): 
            Timestamps of corresponding images.

        images (Dict[str, numpy.ndarray]): 
            Image data prepared into the necessary format; a dictionary. Keys are the site UID, 
            ndarray is the prepared data.

        images_dimensions (Dict[str, Tuple]): 
            The image dimensions.

        data_types (List[str]): 
            The data types for each data object.    
    """

    site_uid_list: List[str]
    timestamps: List[datetime.datetime]
    images: Dict[str, ndarray]
    images_dimensions: Dict[str, Tuple]
    data_types: List[str]

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # set special strings
        unique_dimensions_str = str(list(dict.fromkeys(self.images_dimensions.values()))).replace("[", "").replace("]", "").replace("), (", "),(")
        images_str = "Dict[%d sites of array(dims=%s)]" % (len(self.images.keys()), unique_dimensions_str)
        timestamps_str = "[%d timestamps]" % (len(self.timestamps))

        # return
        return "MosaicData(images=%s, timestamps=%s, site_uid_list=%s)" % (images_str, timestamps_str, self.site_uid_list.__repr__())

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set special strings
        unique_dimensions_str = str(list(dict.fromkeys(self.images_dimensions.values()))).replace("[", "").replace("]", "").replace("), (", "),(")
        images_str = "Dict[%d sites of array(dims=%s)]" % (len(self.images.keys()), unique_dimensions_str)
        timestamps_str = "[%d timestamps]" % (len(self.timestamps))

        # print
        print("MosaicData:")
        print("  %-19s: %s" % ("site_uid_list", self.site_uid_list))
        print("  %-19s: %s" % ("timestamps", timestamps_str))
        print("  %-19s: %s" % ("images", images_str))
        print("  %-19s: %s" % ("images_dimensions", self.images_dimensions))
        print("  %-19s: %s" % ("data_types", self.data_types))


@dataclass
class Mosaic:
    """
    Class representation for a generated mosaic.

    Attributes:
        polygon_data (matplotlib.collections.PolyCollection): 
            Generated polygons containing rendered data.

        cartopy_projection (cartopy.crs.Projection): 
            Cartopy projection to utilize.

        contour_data (Dict[str, List[Any]]): 
            Generated contour data.

        spect_cmap (str): 
            String giving the cmap to use for spect legend.
            
        spect_intensity_scale (Tuple[int]): 
            The min and max values that spectrograph data
            is scaled to in the mosaic, if any is present.
    """
    polygon_data: Union[PolyCollection, List[PolyCollection]]
    cartopy_projection: Projection
    contour_data: Optional[Dict[str, List[Any]]] = None
    spect_cmap: Optional[str] = None
    spect_intensity_scale: Optional[Tuple[int, int]] = None

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        if isinstance(self.polygon_data, list):
            polycollection_str = "[PolyCollection(...), ...]"
        else:
            polycollection_str = "PolyCollection(...)"

        if self.contour_data is not None:
            return "Mosaic(polygon_data=PolyCollection(...), cartopy_projection=Projection(%s), %d Contours)" % (
                self.cartopy_projection.to_string(),
                len(self.contour_data.get("x", [])),
            )
        else:
            return "Mosaic(polygon_data=" + polycollection_str + ", cartopy_projection=Projection(%s))" % (self.cartopy_projection.to_string())

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set special strings
        if isinstance(self.polygon_data, list):
            polycollection_str = "[PolyCollection(...), ...]"
        else:
            polycollection_str = "PolyCollection(...)"
        cartopy_projection_str = "Projection(%s)" % (self.cartopy_projection.to_string())
        if self.contour_data is not None:
            contour_data_str = "%d Contours" % (len(self.contour_data.get("x", [])), )
        else:
            contour_data_str = "None"

        # print
        print("Mosaic:")
        print("  %-23s: %s" % ("polygon_data", polycollection_str))
        print("  %-23s: %s" % ("cartopy_projection", cartopy_projection_str))
        print("  %-23s: %s" % ("contour_data", contour_data_str))
        print("  %-23s: %s" % ("spect_cmap", self.spect_cmap))
        print("  %-23s: %s" % ("spect_intensity_scale", self.spect_intensity_scale))

    def plot(
            self,
            map_extent: Sequence[Union[float, int]],
            figsize: Optional[Tuple[int, int]] = None,
            rayleighs: bool = False,
            max_rayleighs: int = 20000,
            title: Optional[str] = None,
            ocean_color: Optional[str] = None,
            land_color: str = "gray",
            land_edgecolor: str = "#8A8A8A",
            borders_color: str = "#AEAEAE",
            borders_disable: bool = False,
            colorbar_title: Optional[str] = None,  # deprecated in v1.11.0, use cbar_title instead
            cbar_title: Optional[str] = None,
            cbar_colormap: str = "",
            returnfig: bool = False,
            savefig: bool = False,
            savefig_filename: Optional[str] = None,
            savefig_quality: Optional[int] = None) -> Any:
        """
        Generate a plot of the mosaic data. 
        
        Either display it (default behaviour), save it to disk (using the `savefig` parameter), or 
        return the matplotlib plot object for further usage (using the `returnfig` parameter).

        Args:
            map_extent (List[int]): 
                Latitude/longitude range to be visible on the rendered map. This is a list of 4 integers 
                and/or floats, in the order of [min_lon, max_lon, min_lat, max_lat].

            figsize (tuple): 
                The matplotlib figure size to use when plotting. For example `figsize=(14,4)`.

            rayleighs (bool): 
                Set to `True` if the data being plotted is in Rayleighs. Defaults to `False`.

            max_rayleighs (int): 
                Max intensity scale for Rayleighs. Defaults to `20000`.

            title (str): 
                The title to display above the plotted mosaic. Default is no title.

            ocean_color (str): 
                Colour of the ocean. Default is cartopy's default shade of blue. Colours can be supplied
                as a word, or hexcode prefixed with a '#' character (ie. `#55AADD`).
            
            land_color (str): 
                Colour of the land. Default is `gray`. Colours can be supplied as a word, or hexcode 
                prefixed with a '#' character (ie. `#41BB87`).

            land_edgecolor (str): 
                Color of the land edges. Default is `#8A8A8A`. Colours can be supplied as a word, or
                hexcode prefixed with a '#' character.

            borders_color (str): 
                Color of the country borders. Default is `AEAEAE`. Colours can be supplied as a word, or
                hexcode prefixed with a '#' character.
            
            borders_disable (bool): 
                Disbale rendering of the borders. Default is `False`.

            cbar_title (str): 
                Title for the colorbar. Default is no title.

            colorbar_title (str): 
                Deprecated as of v1.10.0. Use 'cbar_title' instead in the exact same way.

            cbar_colormap (str): 
                The matplotlib colormap to use for the plotted color bar. Default is `gray`, unless
                mosaic was created with spectrograph data, in which case defaults to the colormap
                used for spectrograph data..

                Commonly used colormaps are:

                - REGO: `gist_heat`
                - THEMIS ASI: `gray`
                - TREx Blue: `Blues_r`
                - TREx NIR: `gray`
                - TREx RGB: `None`

                A list of all available colormaps can be found on the 
                [matplotlib documentation](https://matplotlib.org/stable/gallery/color/colormap_reference.html).

            returnfig (bool): 
                Instead of displaying the image, return the matplotlib figure object. This allows for further plot 
                manipulation, for example, adding labels or a title in a different location than the default. 
                
                Remember - if this parameter is supplied, be sure that you close your plot after finishing work 
                with it. This can be achieved by doing `plt.close(fig)`. 
                
                Note that this method cannot be used in combination with `savefig`.

            savefig (bool): 
                Save the displayed image to disk instead of displaying it. The parameter savefig_filename is required if 
                this parameter is set to True. Defaults to `False`.

            savefig_filename (str): 
                Filename to save the image to. Must be specified if the savefig parameter is set to True.

            savefig_quality (int): 
                Quality level of the saved image. This can be specified if the savefig_filename is a JPG image. If it
                is a PNG, quality is ignored. Default quality level for JPGs is matplotlib/Pillow's default of 75%.

        Returns:
            The displayed montage, by default. If `savefig` is set to True, nothing will be returned. If `returnfig` is 
            set to True, the plotting variables `(fig, ax)` will be returned.

        Raises:
        """
        # handle deprecation warnings
        if (colorbar_title is not None):
            show_warning("The parameter 'colorbar_title' was deprecated in v1.11.0. Please use 'cbar_title' instead (usage is identical).",
                         stacklevel=1)
            cbar_title = colorbar_title

        # check return mode
        if (returnfig is True and savefig is True):
            raise ValueError("Only one of returnfig or savefig can be set to True")
        if returnfig is True and (savefig_filename is not None or savefig_quality is not None):
            show_warning("The figure will be returned, but a savefig option parameter was supplied. Consider " +
                         "removing the savefig option parameter(s) as they will be ignored.",
                         stacklevel=1)
        elif (savefig is False and (savefig_filename is not None or savefig_quality is not None)):
            show_warning("A savefig option parameter was supplied, but the savefig parameter is False. The " +
                         "savefig option parameters will be ignored.",
                         stacklevel=1)

        # get colormap if there is spectrograph data
        if self.spect_cmap is not None:
            cbar_colormap = self.spect_cmap

        # initialize figure
        fig = plt.figure(figsize=figsize)
        ax = fig.add_axes((0, 0, 1, 1), projection=self.cartopy_projection)
        ax.set_extent(map_extent, crs=cartopy.crs.Geodetic())  # type: ignore

        # add ocean
        #
        # NOTE: we use the default ocean color
        if (ocean_color is not None):
            ax.add_feature(  # type: ignore
                cartopy.feature.OCEAN, facecolor=ocean_color, zorder=0)
        else:
            ax.add_feature(cartopy.feature.OCEAN, zorder=0)  # type: ignore

        # add land
        ax.add_feature(  # type: ignore
            cartopy.feature.LAND, facecolor=land_color, edgecolor=land_edgecolor, zorder=0)

        # add borders
        if (borders_disable is False):
            ax.add_feature(  # type: ignore
                cartopy.feature.BORDERS, edgecolor=borders_color, zorder=0)

        # add polygon data
        #
        # NOTE: it seems that when running this function a second time, the polygon
        # data is not too happy. So to handle this, we plot a copy of the polygon data
        if isinstance(self.polygon_data, list):
            for polygon_data in self.polygon_data:
                ax.add_collection(copy(polygon_data))
        else:
            ax.add_collection(copy(self.polygon_data))

        if self.contour_data is not None:
            for i in range(len(self.contour_data["x"])):
                ax.plot(self.contour_data["x"][i],
                        self.contour_data["y"][i],
                        color=self.contour_data["color"][i],
                        linewidth=self.contour_data["linewidth"][i],
                        linestyle=self.contour_data["linestyle"][i],
                        marker=self.contour_data["marker"][i],
                        zorder=self.contour_data["zorder"][i])

        # set title
        if (title is not None):
            ax.set_title(title)

        # add text
        if (rayleighs is True):
            if isinstance(self.polygon_data, list):
                raise ValueError("Rayleighs Keyword is currently not available for mosaics with multiple sets of data.")

            # Create a colorbar, in Rayleighs, that accounts for the scaling limit we applied
            cbar_ticks = [float(j) / 5. for j in range(0, 6)]
            cbar_ticknames = [str(int(max_rayleighs / 5) * j) for j in range(0, 6)]

            # Any pixels with the max Rayleigh value could be greater than it, so we include the plus sign
            cbar_ticknames[-1] += "+"
            self.polygon_data.set_cmap(cbar_colormap)
            cbar = plt.colorbar(self.polygon_data, shrink=0.5, ticks=cbar_ticks, ax=ax)
            cbar.ax.set_yticklabels(cbar_ticknames)
            plt.text(1.025,
                     0.5,
                     "Intensity (Rayleighs)",
                     fontsize=14,
                     transform=ax.transAxes,
                     va="center",
                     rotation="vertical",
                     weight="bold",
                     style="oblique")

        if (self.spect_cmap) is not None:

            if (self.spect_intensity_scale) is None:
                intensity_max = np.nan
                intensity_min = np.nan
            else:
                intensity_max = self.spect_intensity_scale[1]
                intensity_min = self.spect_intensity_scale[0]

            # Create a colorbar, in Rayleighs, that accounts for the scaling limit we applied
            cbar_ticks = [float(j) / 5. for j in range(0, 6)]
            cbar_ticknames = [str(int((intensity_max / 5) + intensity_min) * j) for j in range(0, 6)]

            # Any specrograph bins with the max intensity value could be greater than it, so we include the plus sign
            cbar_ticknames[-1] += "+"
            if isinstance(self.polygon_data, list):
                self.polygon_data[0].set_cmap(cbar_colormap)
            else:
                self.polygon_data.set_cmap(cbar_colormap)
            if isinstance(self.polygon_data, list):
                cbar = plt.colorbar(self.polygon_data[0], shrink=0.5, ticks=cbar_ticks, ax=ax)
            else:
                cbar = plt.colorbar(self.polygon_data, shrink=0.5, ticks=cbar_ticks, ax=ax)
            cbar.ax.set_yticklabels(cbar_ticknames)
            if (cbar_title is None):
                plt.text(1.025,
                         0.5,
                         "Spectrograph Intensity (Rayleighs)",
                         fontsize=10,
                         transform=ax.transAxes,
                         va="center",
                         rotation="vertical",
                         weight="bold",
                         style="oblique")
            else:
                plt.text(1.025,
                         0.5,
                         cbar_title,
                         fontsize=10,
                         transform=ax.transAxes,
                         va="center",
                         rotation="vertical",
                         weight="bold",
                         style="oblique")

        # save figure or show it
        if (savefig is True):
            # check that filename has been set
            if (savefig_filename is None):
                raise ValueError("The savefig_filename parameter is missing, but required since savefig was set to True.")

            # save the figure
            f_extension = os.path.splitext(savefig_filename)[-1].lower()
            if (".jpg" == f_extension or ".jpeg" == f_extension):
                # check quality setting
                if (savefig_quality is not None):
                    plt.savefig(savefig_filename, quality=savefig_quality, bbox_inches="tight")
                else:
                    plt.savefig(savefig_filename, bbox_inches="tight")
            else:
                if (savefig_quality is not None):
                    # quality specified, but output filename is not a JPG, so show a warning
                    show_warning("The savefig_quality parameter was specified, but is only used for saving JPG files. The " +
                                 "savefig_filename parameter was determined to not be a JPG file, so the quality will be ignored",
                                 stacklevel=1)
                plt.savefig(savefig_filename, bbox_inches="tight")

            # clean up by closing the figure
            plt.close(fig)
        elif (returnfig is True):
            # return the figure and axis objects
            return (fig, ax)
        else:
            # show the figure
            plt.show(fig)

            # cleanup by closing the figure
            plt.close(fig)

        # return
        return None

    def add_geo_contours(self,
                         lats: Optional[Union[ndarray, list]] = None,
                         lons: Optional[Union[ndarray, list]] = None,
                         constant_lats: Optional[Union[float, int, Sequence[Union[float, int]], ndarray]] = None,
                         constant_lons: Optional[Union[float, int, Sequence[Union[float, int]], ndarray]] = None,
                         color: str = "black",
                         linewidth: Union[float, int] = 1,
                         linestyle: str = "solid",
                         marker: str = "",
                         bring_to_front: bool = False):
        """
        Add geographic contours to a mosaic.

        Args:
            lats (ndarray or list): 
                Sequence of geographic latitudes defining a contour.
            
            lons (ndarray or list): 
                Sequence of geographic longitudes defining a contour.

            constant_lats (float, int, or Sequence): 
                Geographic Latitude(s) at which to add line(s) of constant latitude.
            
            constant_lons (float, int, or Sequence): 
                Geographic Longitude(s) at which to add line(s) of constant longitude.

            color (str): 
                The matplotlib color used for the contour(s).

            linewidth (float or int): 
                The contour thickness.
            
            linestyle (str): 
                The matplotlib linestyle used for the contour(s).

            marker (str): 
                The matplotlib marker used for the contour(s).

            bring_to_front (bool): 
                Plots the contour on top of all other currently plotted objects.

        Returns:
            The object's contour_data parameter is populated appropriately.

        Raises:
            ValueError: issues encountered with supplied parameters.
        """
        # Make sure some form of lat/lon is provided
        if (constant_lats is None) and (constant_lons is None) and (lats is None) and (lons is None):
            raise ValueError("No latitudes or longitudes provided.")

        # If manually passing in lats & lons, make sure both are provided
        if (lats is not None or lons is not None) and (lats is None or lons is None):
            raise (ValueError("Manually supplying contour requires both lats and lons."))

        # Check that color exists in matplotlib
        if color not in matplotlib.colors.CSS4_COLORS:
            raise ValueError(f"Color '{color}' not recognized by matplotlib.")

        # Check that linestyle is valid
        if linestyle not in ["-", "--", "-.", ":", "solid", "dashed", "dashdot", "dotted"]:
            raise ValueError(f"Linestyle '{linestyle}' not recognized by matplotlib.")

        # Check that linewidth is valid
        if linewidth <= 0:
            raise ValueError("Linewidth must be greater than zero.")

        # Check that marker is valid
        if marker not in ["", "o", ".", "p", "*", "x", "+", "X"]:
            raise ValueError(f"Marker '{marker}' is not currently supported.")

        # Convert numerics to lists if necessary
        if constant_lats is not None:
            if isinstance(constant_lats, (float, int)):
                constant_lats = [constant_lats]
        if constant_lons is not None:
            if isinstance(constant_lons, (float, int)):
                constant_lons = [constant_lons]

        # Initialize contour data dict if it doesn't exist yet
        if self.contour_data is None:
            self.contour_data = {"x": [], "y": [], "color": [], "linewidth": [], "linestyle": [], "marker": [], "zorder": []}

        # Obtain the mosaic's projection
        source_proj = pyproj.CRS.from_user_input(cartopy.crs.Geodetic())
        mosaic_proj = pyproj.CRS.from_user_input(self.cartopy_projection)
        transformer = pyproj.Transformer.from_crs(source_proj, mosaic_proj, always_xy=True)

        # First handling manually supplied lat/lon arrays
        if (lats is not None) and (lons is not None):
            # Convert lists to ndarrays if necessary
            if isinstance(lats, list):
                lats = np.array(lats)
            if isinstance(lons, list):
                lons = np.array(lons)

            if len(lats) != len(lons):
                raise ValueError("Lat/Lon data must be of the same size.")

            # Create specified contour from geographic coords
            x, y = transformer.transform(lons, lats)
            # Add contour to dict, along with color and linewidth
            self.contour_data["x"].append(x)
            self.contour_data["y"].append(y)
            self.contour_data["color"].append(color)
            self.contour_data["linewidth"].append(linewidth)
            self.contour_data["linestyle"].append(linestyle)
            self.contour_data["marker"].append(marker)
            self.contour_data["zorder"].append(int(bring_to_front))

        # Next handling lines of constant latitude
        if constant_lats is not None:
            # Generate longitudinal domain of the lat line (full globe)
            lon_domain = np.arange(-180, 180 + 0.2, 0.2)

            # Iterate through all lines of constant lat requested
            for lat in constant_lats:
                # Create line of constant lat
                const_lat_x, const_lat_y = (lon_domain, lon_domain * 0 + lat)
                sort_idx = np.argsort(const_lat_x)
                const_lat_y = const_lat_y[sort_idx]
                const_lat_x = const_lat_x[sort_idx]
                const_lat_x, const_lat_y = transformer.transform(const_lat_x, const_lat_y)

                # Add contour to dict, along with color and linewidth
                self.contour_data["x"].append(const_lat_x)
                self.contour_data["y"].append(const_lat_y)
                self.contour_data["color"].append(color)
                self.contour_data["linewidth"].append(linewidth)
                self.contour_data["linestyle"].append(linestyle)
                self.contour_data["marker"].append(marker)
                self.contour_data["zorder"].append(int(bring_to_front))

        # Now handling lines of constant longitude
        if constant_lons is not None:
            # Generate latitudinal domain of the lon line (full globe)
            lat_domain = np.arange(-90, 90 + 0.1, 0.1)

            # Iterate through all lines of constant lon requested
            for lon in constant_lons:
                # Create line of constant lon and add to dict
                const_lon_x, const_lon_y = (lat_domain * 0 + lon, lat_domain)
                sort_idx = np.argsort(const_lon_y)
                const_lon_x = const_lon_x[sort_idx]
                const_lon_y = const_lon_y[sort_idx]
                const_lon_x, const_lon_y = transformer.transform(const_lon_x, const_lon_y)

                # Add contour to dict, along with color and linewidth
                self.contour_data["x"].append(const_lon_x)
                self.contour_data["y"].append(const_lon_y)
                self.contour_data["color"].append(color)
                self.contour_data["linewidth"].append(linewidth)
                self.contour_data["linestyle"].append(linestyle)
                self.contour_data["marker"].append(marker)
                self.contour_data["zorder"].append(int(bring_to_front))

    def add_mag_contours(self,
                         timestamp: datetime.datetime,
                         constant_lats: Optional[Union[float, int, Sequence[Union[float, int]], ndarray]] = None,
                         constant_lons: Optional[Union[float, int, Sequence[Union[float, int]], ndarray]] = None,
                         lats: Optional[Union[ndarray, list]] = None,
                         lons: Optional[Union[ndarray, list]] = None,
                         color: str = "black",
                         linewidth: Union[float, int] = 1,
                         linestyle: str = "solid",
                         marker: str = "",
                         bring_to_front: bool = False):
        """
        Add geomagnetic contours to a mosaic.

        Args:
            timestamp (datetime.datetime): 
                The timestamp used in computing AACGM coordinates.

            lats (ndarray or list): 
                Sequence of geomagnetic latitudes defining a contour.
            
            lons (ndarray or list): 
                Sequence of geomagnetic longitudes defining a contour.

            constant_lats (float, int, Sequence): 
                Geomagnetic latitude(s) at which to add contour(s) of constant latitude.
            
            constant_lons (float, int, Sequence): 
                Geomagnetic longitude(s) at which to add contours(s) of constant longitude.

            color (str): 
                The matplotlib color used for the contour(s).

            linewidth (float or int): 
                The contour thickness.

            linestyle (str): 
                The matplotlib linestyle used for the contour(s).

            marker (str): 
                The matplotlib marker used for the contour(s).

            bring_to_front (bool): 
                Plots the contour on top of all other currently plotted objects.

        Returns:
            The object's contour_data parameter is populated appropriately.

        Raises:
            ValueError: issues encountered with supplied parameters
        """
        # Make sure some form of lat/lon is provided
        if (constant_lats is None) and (constant_lons is None) and (lats is None) and (lons is None):
            raise ValueError("No latitudes or longitudes provided.")

        # If manually passing in lats & lons, make sure both are provided
        if (lats is not None or lons is not None) and (lats is None or lons is None):
            raise ValueError("Manually supplying contour requires both lats and lons.")

        # Check that color exists in matplotlib
        if color not in matplotlib.colors.CSS4_COLORS:
            raise ValueError(f"Color '{color}' not recognized by matplotlib.")

        # Check that linestyle is valid
        if linestyle not in ["-", "--", "-.", ":", "solid", "dashed", "dashdot", "dotted"]:
            raise ValueError(f"Linestyle '{linestyle}' not recognized by matplotlib.")

        # Check that linewidth is valid
        if linewidth <= 0:
            raise ValueError("linewidth must be greater than zero.")

        # Convert numerics to lists if necessary
        if constant_lats is not None:
            if isinstance(constant_lats, (float, int)):
                constant_lats = [constant_lats]
        if constant_lons is not None:
            if isinstance(constant_lons, (float, int)):
                constant_lons = [constant_lons]

        # Initialize contour data dict if it doesn't exist yet
        if self.contour_data is None:
            self.contour_data = {"x": [], "y": [], "color": [], "linewidth": [], "linestyle": [], "marker": [], "zorder": []}

        # Obtain the mosaic's projection
        source_proj = pyproj.CRS.from_user_input(cartopy.crs.Geodetic())
        mosaic_proj = pyproj.CRS.from_user_input(self.cartopy_projection)
        transformer = pyproj.Transformer.from_crs(source_proj, mosaic_proj, always_xy=True)

        # First handling manually supplied lat/lon arrays
        if (lats is not None) and (lons is not None):
            # Convert lists to ndarrays if necessary
            if isinstance(lats, list):
                lats = np.array(lats)
            if isinstance(lons, list):
                lons = np.array(lons)

            if len(lats) != len(lons):
                raise ValueError("Lat/Lon data must be of the same size.")

            # Create specified contour from magnetic coords
            y, x, alt = aacgmv2.convert_latlon_arr(lats, lons, lats * 0.0, timestamp, method_code="A2G")
            x, y = transformer.transform(x, y)

            # Add contour to dict, along with color and linewidth
            self.contour_data["x"].append(x)
            self.contour_data["y"].append(y)
            self.contour_data["color"].append(color)
            self.contour_data["linewidth"].append(linewidth)
            self.contour_data["linestyle"].append(linestyle)
            self.contour_data["marker"].append(marker)
            self.contour_data["zorder"].append(int(bring_to_front))

        # Next handling lines of constant latitude
        if constant_lats is not None:
            # Generate longitudinal domain of the lat line (full globe)
            lon_domain = np.arange(-180, 180 + 0.2, 0.2)

            # iterate through all lines of constant lat requested
            for lat in constant_lats:
                # Create line of constant lat from magnetic coords
                const_lat_x, const_lat_y = (lon_domain, lon_domain * 0 + lat)
                const_lat_y, const_lat_x, alt = aacgmv2.convert_latlon_arr(const_lat_y, const_lat_x, const_lat_x * 0.0, timestamp, method_code="A2G")
                sort_idx = np.argsort(const_lat_x)
                const_lat_y = const_lat_y[sort_idx]
                const_lat_x = const_lat_x[sort_idx]
                const_lat_x, const_lat_y = transformer.transform(const_lat_x, const_lat_y)

                # Add contour to dict, along with color and linewidth
                self.contour_data["x"].append(const_lat_x)
                self.contour_data["y"].append(const_lat_y)
                self.contour_data["color"].append(color)
                self.contour_data["linewidth"].append(linewidth)
                self.contour_data["linestyle"].append(linestyle)
                self.contour_data["marker"].append(marker)
                self.contour_data["zorder"].append(int(bring_to_front))

        # Now handling lines of constant longitude
        if constant_lons is not None:
            # Generate latitudinal domain of the lon line (full globe)
            lat_domain = np.arange(-90, 90 + 0.1, 0.1)

            # iterate through all lines of constant lon requested
            for lon in constant_lons:
                # Create line of constant lon from magnetic coords
                const_lon_x, const_lon_y = (lat_domain * 0 + lon, lat_domain)
                const_lon_y, const_lon_x, alt = aacgmv2.convert_latlon_arr(const_lon_y, const_lon_x, const_lon_x * 0.0, timestamp, method_code="A2G")
                sort_idx = np.argsort(const_lon_y)
                const_lon_x = const_lon_x[sort_idx]
                const_lon_y = const_lon_y[sort_idx]
                const_lon_x, const_lon_y = transformer.transform(const_lon_x, const_lon_y)

                # Add contour to dict, along with color and linewidth
                self.contour_data["x"].append(const_lon_x)
                self.contour_data["y"].append(const_lon_y)
                self.contour_data["color"].append(color)
                self.contour_data["linewidth"].append(linewidth)
                self.contour_data["linestyle"].append(linestyle)
                self.contour_data["marker"].append(marker)
                self.contour_data["zorder"].append(int(bring_to_front))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import requests
import click
import pyaurorax
from .search import search_group

# default context settings
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class Config(object):

    def __init__(self, verbose=False, api_key=None, api_base_url=None):
        self.verbose = verbose
        self.aurorax = pyaurorax.PyAuroraX()
        if (api_key is not None):
            self.aurorax.api_key = api_key
        if (api_base_url is not None):
            self.aurorax.api_base_url = api_base_url


def __test_connectivity(aurorax):
    # make request
    click.echo("Checking connectivity to %s ...\n" % (aurorax.api_base_url))
    try:
        r = requests.get(aurorax.api_base_url, timeout=aurorax.api_timeout)
    except requests.RequestException as e:  # pragma: nocover-ok
        click.echo("Error connecting to AuroraX API: %s" % (str(e)))
        sys.exit(1)

    # check status code
    if (r.status_code == 200):
        click.echo("Connectivity to the AuroraX API looks good!")
    else:
        click.echo("Error connecting to AuroraX API, got a %d response" % (r.status_code))
        sys.exit(1)


@click.group(invoke_without_command=True)
@click.version_option(version="1.17.0")
@click.option("--api-key", type=str, help="Specify an API key")
@click.option("--api-base-url", type=str, help="Set the AuroraX API base URL")
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
@click.option("--test-connectivity", is_flag=True, help="Test connectivity to AuroraX API")
@click.pass_context
def cli(ctx, api_key, api_base_url, verbose, test_connectivity):
    """
    Welcome to the PyAuroraX CLI program!

    This program is meant to provide an easy interface with AuroraX
    from the command line. It uses the PyAuroraX library behind the scenes.
    """
    # set config
    ctx.obj = Config(verbose=verbose, api_key=api_key, api_base_url=api_base_url)

    # evaluate options
    if (ctx.invoked_subcommand is None):
        if (test_connectivity is True):
            # evaluate --test-connectivity
            __test_connectivity(ctx.obj.aurorax)
        else:
            # no options called, output the help
            click.echo("""Welcome to the PyAuroraX CLI program!

This program is meant to provide an easy interface with AuroraX from the
command line.

To learn more about usage, type:

  $> aurorax-cli --help
""")
    else:
        # subcommand was called, move on to that
        pass


# add sub commands
cli.add_command(search_group)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class representation for a montage.
"""

import os
import datetime
import matplotlib.pyplot as plt
import numpy as np
from dataclasses import dataclass
from typing import List, Optional, Tuple, Any
from ..._util import show_warning


@dataclass
class Montage:
    """
    Class representation for a montage

    Attributes:
        data (numpy.ndarray): 
            The derived montage data.

        timestamp (List[datetime.datetime]): 
            Timestamps corresponding to each montage image.
    """

    def __init__(self, data: np.ndarray, timestamp: List[datetime.datetime], n_channels: int):
        # public vars
        self.data = data
        self.timestamp = timestamp

        # private vars
        self.__n_channels = n_channels

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        data_str = "array(dims=%s, dtype=%s)" % (self.data.shape, self.data.dtype)
        timestamp_str = "[%d datetime objects]" % (len(self.timestamp))

        return "Montage(data=%s, timestamp=%s)" % (data_str, timestamp_str)

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set special strings
        data_str = "array(dims=%s, dtype=%s)" % (self.data.shape, self.data.dtype)
        timestamp_str = "[%d datetime objects]" % (len(self.timestamp))

        # print
        print("Montage:")
        print("  %-11s: %s" % ("data", data_str))
        print("  %-11s: %s" % ("timestamp", timestamp_str))

    def plot(self,
             rows: int,
             cols: int,
             timestamps_display: bool = True,
             timestamps_format: str = "%Y-%m-%d %H:%M:%S",
             timestamps_fontsize: int = 11,
             figsize: Optional[Tuple[int, int]] = None,
             title: Optional[str] = None,
             cmap: Optional[str] = None,
             returnfig: bool = False,
             savefig: bool = False,
             savefig_filename: Optional[str] = None,
             savefig_quality: Optional[int] = None) -> Any:
        """
        Generate a plot of the montage data. 
        
        Either display it (default behaviour), save it to disk (using the `savefig` parameter), or 
        return the matplotlib plot object for further usage (using the `returnfig` parameter).

        Args:
            rows (int): 
                The number of rows to use when displaying the images as a montage. The product of 
                the `rows` and `cols` parameters must match the number of images supplied when creating 
                the `Montage` object. If not, a ValueError will be raised. This parameter is required.

            cols (int): 
                The number of columns to use when displaying the images as a montage. The product of 
                the `rows` and `cols` parameters must match the number of images supplied when creating 
                the `Montage` object. If not, a ValueError will be raised. This parameter is required.

            timestamps_display (bool): 
                Display the timestamps on each montage image. Defaults to True. This parameter is
                optional.

            timestamps_format (str): 
                The format of the timestamps when being displayed. This is the same format string
                as when using the `strftime()` function for a `datetime` object. Default format 
                string is `%Y-%m-%d %H:%M:%S`. Refer to the 
                [Python documentation](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes)
                for further information. This parameter is optional.
            
            timestamps_fontsize (int): 
                The font size for the displayed timestamps. Default is size 11. This parameter is 
                optional.

            figsize (tuple): 
                The matplotlib figure size to use when plotting. For example `figsize=(14,4)`.

            title (str): 
                The title to display above the plotted montage. Default is no title.

            cmap (str): 
                The matplotlib colormap to use.

                Commonly used colormaps are:

                - REGO: `gist_heat`
                - THEMIS ASI: `gray`
                - TREx Blue: `Blues_r`
                - TREx NIR: `gray`
                - TREx RGB: `None`

                A list of all available colormaps can be found on the 
                [matplotlib documentation](https://matplotlib.org/stable/gallery/color/colormap_reference.html).
            
            returnfig (bool): 
                Instead of displaying the image, return the matplotlib figure object. This allows for further plot 
                manipulation, for example, adding labels or a title in a different location than the default. 
                
                Remember - if this parameter is supplied, be sure that you close your plot after finishing work 
                with it. This can be achieved by doing `plt.close(fig)`. 
                
                Note that this method cannot be used in combination with `savefig`.

            savefig (bool): 
                Save the displayed image to disk instead of displaying it. The parameter savefig_filename is required if 
                this parameter is set to True. Defaults to `False`.

            savefig_filename (str): 
                Filename to save the image to. Must be specified if the savefig parameter is set to True.

            savefig_quality (int): 
                Quality level of the saved image. This can be specified if the savefig_filename is a JPG image. If it
                is a PNG, quality is ignored. Default quality level for JPGs is matplotlib/Pillow's default of 75%.

        Returns:
            The displayed montage, by default. If `savefig` is set to True, nothing will be returned. If `returnfig` is 
            set to True, the plotting variables `(fig, ax)` will be returned.

        Raises:
            ValueError: issues with the y-axis choice, or rows/cols choice.
        """
        # check rows and cols
        if (rows * cols != self.data.shape[-1]):
            raise ValueError(("Invalid choice of rows and columns. The %d images in this Montage object " +
                              "cannot be organized into a %dx%d grid (%d * %d != %d)") % (
                                  self.data.shape[-1],
                                  rows,
                                  cols,
                                  rows,
                                  cols,
                                  self.data.shape[-1],
                              ))

        # check return mode
        if (returnfig is True and savefig is True):
            raise ValueError("Only one of returnfig or savefig can be set to True")
        if (returnfig is True and (savefig_filename is not None or savefig_quality is not None)):
            show_warning("The figure will be returned, but a savefig option parameter was supplied. Consider " +
                         "removing the savefig option parameter(s) as they will be ignored.",
                         stacklevel=1)
        elif (savefig is False and (savefig_filename is not None or savefig_quality is not None)):
            show_warning("A savefig option parameter was supplied, but the savefig parameter is False. The " +
                         "savefig option parameters will be ignored.",
                         stacklevel=1)

        # init figure
        fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

        # set title
        if (title is not None):
            fig.suptitle(title)

        # for each image
        for ax, i in zip(axs.flat, range(0, len(self.timestamp))):  # type: ignore
            if (self.__n_channels == 1):
                # single channel
                ax.imshow(self.data[:, :, i], cmap=cmap, origin="lower", interpolation="nearest")
            elif (self.__n_channels == 3):
                # single channel
                ax.imshow(self.data[:, :, :, i], cmap=cmap, origin="lower", interpolation="nearest")
            else:  # pragma: nocover
                raise ValueError("Can only plot 3 or 4 dimensional data (series of single-channel or RGB mages), but found data of shape %s" %
                                 (self.data.shape))

            ax.set_axis_off()

            # show timestamp
            if (timestamps_display is True):
                ax.text(
                    int(np.floor(self.data.shape[1] / 2.)),
                    5,
                    self.timestamp[i].strftime(timestamps_format),
                    ha="center",
                    fontsize=timestamps_fontsize,
                )
        plt.tight_layout(h_pad=0, w_pad=0)

        # save figure or show it
        if (savefig is True):
            # check that filename has been set
            if (savefig_filename is None):
                raise ValueError("The savefig_filename parameter is missing, but required since savefig was set to True.")

            # save the figure
            f_extension = os.path.splitext(savefig_filename)[-1].lower()
            if (".jpg" == f_extension or ".jpeg" == f_extension):
                # check quality setting
                if (savefig_quality is not None):  # pragma: nocover
                    plt.savefig(savefig_filename, quality=savefig_quality, bbox_inches="tight")
                else:
                    plt.savefig(savefig_filename, bbox_inches="tight")
            else:
                if (savefig_quality is not None):
                    # quality specified, but output filename is not a JPG, so show a warning
                    show_warning("The savefig_quality parameter was specified, but is only used for saving JPG files. The " +
                                 "savefig_filename parameter was determined to not be a JPG file, so the quality will be ignored",
                                 stacklevel=1)
                plt.savefig(savefig_filename, bbox_inches="tight")

            # clean up by closing the figure
            plt.close(fig)
        elif (returnfig is True):
            # return the figure and axis objects
            return (fig, axs)
        else:
            # show the figure
            plt.show(fig)

            # cleanup by closing the figure
            plt.close(fig)

        # return
        return None



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from ...data.ucalgary import Calibration


def perform_dark_frame_correction(images: np.ndarray, size: int) -> np.ndarray:
    """
    This method will perform a dark frame correction by subtracting an average
    of a bottom corner grid from the image (ie. 4x4.).

    NOTE: This is an internal-only used function. It is not publicly exposed.
    """
    # init
    original_dtype = images.dtype
    if (len(images.shape) == 2):
        # only a single image, add an axis
        images = images[:, :, np.newaxis]

    # extract NxN box from lower left corner, compute means
    dark_boxes = images[0:size, 0:size, :]
    dark_means = np.mean(dark_boxes, axis=(0, 1))
    dark_means = dark_means.astype(np.int32)

    # apply
    #
    # NOTE: we convert to int32 to avoid lower-bound rollover and allow
    # for negative numbers, which we then cast to 0 and convert back to
    # the native dtype.
    #
    # TODO: there's definitely some optimization possible here
    images = images.astype(np.int32)
    for i in range(0, images.shape[2]):
        images[:, :, i] = images[:, :, i] - dark_means[i]
    images[np.where(images < 0)] = 0
    images = images.astype(original_dtype)

    # remove last axis if there's only one image
    if (images.shape[2] == 1):
        images = np.squeeze(images, axis=2)

    # return
    return images


def perform_flatfield_calibration(images: np.ndarray, cal_flatfield: Calibration) -> np.ndarray:
    # add axis if it's a single image
    if (len(images.shape) == 2):
        # only a single image, add an axis
        images = images[:, :, np.newaxis]

    # for each image, apply multiplier
    for i in range(0, images.shape[-1]):
        images[:, :, i] = images[:, :, i] * cal_flatfield.flat_field_multiplier

    # remove last axis if there's only one image
    if (images.shape[2] == 1):
        images = np.squeeze(images, axis=2)

    # return
    return images


def perform_rayleighs_correction(images: np.ndarray, cal_rayleighs: Calibration, exposure_length_sec: float) -> np.ndarray:
    # add axis if it's a single image
    if (len(images.shape) == 2):
        # only a single image, add an axis
        images = images[:, :, np.newaxis]

    # convert image array to a float array
    images = images.astype(np.float32)

    # for each images, apply rayleighs conversion
    for i in range(0, images.shape[-1]):
        images[:, :, i] = images[:, :, i] * cal_rayleighs.rayleighs_perdn_persecond / float(exposure_length_sec)

    # remove last axis if there's only one image
    if (images.shape[2] == 1):
        images = np.squeeze(images, axis=2)

    # return
    return images



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pprint
import click
import humanize
import pyaurorax
import textwrap
from texttable import Texttable

# globals
SUPPORTED_SOURCE_TYPES = [
    pyaurorax.search.SOURCE_TYPE_EVENT_LIST,
    pyaurorax.search.SOURCE_TYPE_GROUND,
    pyaurorax.search.SOURCE_TYPE_HEO,
    pyaurorax.search.SOURCE_TYPE_LEO,
    pyaurorax.search.SOURCE_TYPE_LUNAR,
]
ALLOWED_FORMATS = [
    pyaurorax.search.FORMAT_BASIC_INFO,
    pyaurorax.search.FORMAT_BASIC_INFO_WITH_METADATA,
    pyaurorax.search.FORMAT_IDENTIFIER_ONLY,
    pyaurorax.search.FORMAT_FULL_RECORD,
]


def __print_stats(stats):
    if (stats.ephemeris_count == 0):  # pragma: nocover
        click.echo("Ephemeris Stats:\t0 records")
    else:
        click.echo("Ephemeris Stats:\t%s (%s to %s)" % (
            humanize.intcomma(stats.ephemeris_count),
            stats.earliest_ephemeris_loaded.strftime("%Y-%m-%dT%H:%M"),
            stats.latest_ephemeris_loaded.strftime("%Y-%m-%dT%H:%M"),
        ))
    if (stats.data_product_count == 0):
        click.echo("Data Product Stats:\t0 records")
    else:
        click.echo("Data Product Stats:\t%s (%s to %s)" % (
            humanize.intcomma(stats.data_product_count),
            stats.earliest_data_product_loaded.strftime("%Y-%m-%dT%H:%M"),
            stats.latest_data_product_loaded.strftime("%Y-%m-%dT%H:%M"),
        ))


def __print_metadata_schema_table(ephemeris_schema=[], data_product_schema=[]):
    # init
    wrap_threshold = 40

    # set table lists
    table_schemas = []
    table_allowed_values = []
    table_data_types = []
    table_descriptions = []
    table_additional_descriptions = []
    table_field_names = []
    table_searchable = []
    for item in ephemeris_schema:
        table_schemas.append("ephemeris")
        table_field_names.append(item["field_name"])
        table_data_types.append(item["data_type"])
        table_allowed_values.append('\n'.join(textwrap.wrap(str(item["allowed_values"]), wrap_threshold)))
        table_descriptions.append('\n'.join(textwrap.wrap(item["description"], wrap_threshold)))
        if ("additional_description" in item and item["additional_description"] is not None):
            table_additional_descriptions.append('\n'.join(textwrap.wrap(item["additional_description"], wrap_threshold)))
        else:
            table_additional_descriptions.append("-")
        if ("searchable" in item):
            table_searchable.append(item["searchable"])
        else:
            table_searchable.append("-")  # pragma: nocover
    for item in data_product_schema:
        table_schemas.append("data_product")
        table_field_names.append(item["field_name"])
        table_data_types.append(item["data_type"])
        table_allowed_values.append('\n'.join(textwrap.wrap(str(item["allowed_values"]), wrap_threshold)))
        table_descriptions.append('\n'.join(textwrap.wrap(item["description"], wrap_threshold)))
        if ("additional_description" in item):
            table_additional_descriptions.append('\n'.join(textwrap.wrap(item["additional_description"], wrap_threshold)))
        else:
            table_additional_descriptions.append("-")  # pragma: nocover
        if ("searchable" in item):
            table_searchable.append(item["searchable"])
        else:
            table_searchable.append("-")  # pragma: nocover

    # set header values
    table_headers = ["Schema", "Field Name", "Data Type", "Allowed Values", "Searchable", "Description", "Additional Description"]

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_field_names)):
        table.add_row([
            table_schemas[i], table_field_names[i], table_data_types[i], table_allowed_values[i], table_searchable[i], table_descriptions[i],
            table_additional_descriptions[i]
        ])
    click.echo(table.draw())


def __print_single_data_source(ds, format, include_stats):
    if (format == pyaurorax.search.FORMAT_IDENTIFIER_ONLY):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        if (include_stats is True):
            __print_stats(ds.stats)
    elif (format == pyaurorax.search.FORMAT_BASIC_INFO):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
        if (include_stats is True):
            __print_stats(ds.stats)
    elif (format == pyaurorax.search.FORMAT_BASIC_INFO_WITH_METADATA):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
        if (ds.metadata == {}):
            click.echo("Metadata:\t\t%s" % (ds.metadata))
        else:
            click.echo("Metadata:\n%s" % (pprint.pformat(ds.metadata)))  # pragma: nocover
        if (include_stats is True):
            __print_stats(ds.stats)
    elif (format == pyaurorax.search.FORMAT_FULL_RECORD):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
        click.echo("Owner:\t\t\t%s" % (ds.owner))
        click.echo("Maintainers:\t\t%s" % (ds.maintainers))
        if (ds.metadata == {}):
            click.echo("Metadata:\t\t%s" % (ds.metadata))
        else:
            click.echo("Metadata:\n%s" % (pprint.pformat(ds.metadata)))  # pragma: nocover
        if (include_stats is True):
            __print_stats(ds.stats)
        if (ds.ephemeris_metadata_schema == []):
            click.echo("Ephemeris Schema:\t[]")
        else:
            click.echo("Ephemeris Schema:\tsee below table")
        if (ds.data_product_metadata_schema == []):
            click.echo("Data Product Schema:\t[]\n")
        else:
            click.echo("Data Product Schema:\tsee below table\n")
        __print_metadata_schema_table(ephemeris_schema=ds.ephemeris_metadata_schema, data_product_schema=ds.data_product_metadata_schema)


def __print_sources_table(sources, order, show_owner, include_stats):
    # set table lists
    table_identifiers = []
    table_programs = []
    table_platforms = []
    table_instrument_types = []
    table_source_types = []
    table_display_names = []
    table_owners = []
    table_stats_ephemeris = []
    table_stats_data_products = []
    sum_ephemeris_count = 0
    sum_data_product_count = 0
    for source in sources:
        table_identifiers.append(source.identifier)
        table_programs.append(source.program)
        table_platforms.append(source.platform)
        table_instrument_types.append(source.instrument_type)
        table_source_types.append(source.source_type)
        table_display_names.append(source.display_name)
        table_owners.append(source.owner)
        if (source.stats is not None):
            if (source.stats.earliest_ephemeris_loaded is not None and source.stats.latest_ephemeris_loaded is not None):
                table_stats_ephemeris.append(
                    "%s records (%s to %s)" %
                    (humanize.intcomma(source.stats.ephemeris_count), source.stats.earliest_ephemeris_loaded.strftime("%Y-%m-%dT%H:%M"),
                     source.stats.latest_ephemeris_loaded.strftime("%Y-%m-%dT%H:%M")))
                sum_ephemeris_count += source.stats.ephemeris_count
            else:  # pragma: nocover
                table_stats_ephemeris.append("%s records" % (humanize.intcomma(source.stats.ephemeris_count)))
            if (source.stats.earliest_data_product_loaded is not None and source.stats.latest_data_product_loaded is not None):
                table_stats_data_products.append(
                    "%s records (%s to %s)" %
                    (humanize.intcomma(source.stats.data_product_count), source.stats.earliest_data_product_loaded.strftime("%Y-%m-%dT%H:%M"),
                     source.stats.latest_data_product_loaded.strftime("%Y-%m-%dT%H:%M")))
                sum_data_product_count += source.stats.data_product_count
            else:
                table_stats_data_products.append("%s records" % (humanize.intcomma(source.stats.data_product_count)))
        else:
            table_stats_ephemeris.append("-")
            table_stats_data_products.append("-")

    # set header values
    table_headers = ["Identifier", "Display Name", "Program", "Platform", "Instrument Type", "Source Type"]
    if (show_owner is True):
        table_headers.append("Owner")
    if (include_stats is True):
        table_headers.append("Ephemeris stats")
        table_headers.append("Data Products stats")
    for i in range(0, len(table_headers)):
        if (table_headers[i].lower().replace(' ', '_') == order):
            table_headers[i] += " " + "\u2193"

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_identifiers)):
        if (show_owner is True and include_stats is True):
            table.add_row([
                table_identifiers[i], table_display_names[i], table_programs[i], table_platforms[i], table_instrument_types[i], table_source_types[i],
                table_owners[i], table_stats_ephemeris[i], table_stats_data_products[i]
            ])
        elif (show_owner is True):
            table.add_row([
                table_identifiers[i], table_display_names[i], table_programs[i], table_platforms[i], table_instrument_types[i], table_source_types[i],
                table_owners[i]
            ])
        elif (include_stats is True):
            table.add_row([
                table_identifiers[i], table_display_names[i], table_programs[i], table_platforms[i], table_instrument_types[i], table_source_types[i],
                table_stats_ephemeris[i], table_stats_data_products[i]
            ])
        else:
            table.add_row([
                table_identifiers[i], table_display_names[i], table_programs[i], table_platforms[i], table_instrument_types[i], table_source_types[i]
            ])
    click.echo(table.draw())

    # show stats sums
    if (include_stats is True):
        print("\nTotal Ephemeris records:\t%s" % (humanize.intcomma(sum_ephemeris_count)))
        print("Total Data Product records:\t%s\n" % (humanize.intcomma(sum_data_product_count)))


@click.group("sources", help="Interact with data sources")
def sources_group():
    pass


@sources_group.command("list", short_help="List data sources")
@click.option("--program", type=str, help="Filter using program")
@click.option("--platform", type=str, help="Filter using platform")
@click.option("--instrument-type", type=str, help="Filter using instrument type")
@click.option("--source-type", type=click.Choice(SUPPORTED_SOURCE_TYPES), help="Filter using source type")
@click.option("--owner", type=str, help="Filter using an owner")
@click.option("--order",
              type=click.Choice(["identifier", "program", "platform", "instrument_type", "display_name", "owner"]),
              default="identifier",
              show_default=True,
              help="Order results using a certain column")
@click.option("--include-stats", is_flag=True, help="Include additional information about data sources")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def list(config, program, platform, instrument_type, source_type, owner, order, include_stats, reversed_):
    """
    List data sources using the options to filter as desired
    """
    # get data sources
    try:
        sources = config.aurorax.search.sources.get_using_filters(program=program,
                                                                  platform=platform,
                                                                  instrument_type=instrument_type,
                                                                  source_type=source_type,
                                                                  owner=owner,
                                                                  order=order,
                                                                  include_stats=include_stats)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # reverse
    if (reversed_ is True):
        sources = reversed(sources)

    # decide if we want to show the owner
    show_owner = False
    if (owner is not None or order == "owner"):
        show_owner = True

    # print the table
    __print_sources_table(sources, order, show_owner, include_stats)


@sources_group.command("search", short_help="Search for data sources")
@click.option("--programs", type=str, help="Search for program (comma separate for multiple values)")
@click.option("--platforms", type=str, help="Search for platform (comma separate for multiple values)")
@click.option("--instrument-types", type=str, help="Search for instrument type (comma separate for multiple values)")
@click.option("--order",
              type=click.Choice(["identifier", "program", "platform", "instrument_type", "display_name", "owner"]),
              default="identifier",
              show_default=True,
              help="Order results using a certain column")
@click.option("--include-stats", is_flag=True, help="Include additional information about data sources")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def search(config, programs, platforms, instrument_types, order, include_stats, reversed_):
    """
    Search for data sources using the options to filter as desired. Unlike
    the 'list' command filters, this command supports multiple programs,
    platforms, or instrument types (using commas).
    """
    # set programs values
    parsed_programs = []
    if (programs is None):
        pass
    elif (',' in programs):
        for p in programs.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_programs.append(p)
    else:
        parsed_programs = [programs]

    # set platforms values
    parsed_platforms = []
    if (platforms is None):
        pass
    elif (',' in platforms):
        for p in platforms.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_platforms.append(p)
    else:
        parsed_platforms = [platforms]

    # set instrument_types values
    parsed_instrument_types = []
    if (instrument_types is None):
        pass
    elif (',' in instrument_types):
        for p in instrument_types.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_instrument_types.append(p)
    else:
        parsed_instrument_types = [instrument_types]

    # search for data sources
    try:
        sources = config.aurorax.search.sources.search(programs=parsed_programs,
                                                       platforms=parsed_platforms,
                                                       instrument_types=parsed_instrument_types,
                                                       order=order,
                                                       include_stats=include_stats)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # reverse
    if (reversed_ is True):
        sources = reversed(sources)

    # decide if we want to show the owner
    show_owner = False
    if (order == "owner"):
        show_owner = True

    # print the table
    __print_sources_table(sources, order, show_owner, include_stats)


@sources_group.command("get", short_help="Get a single data source")
@click.argument("program", type=str)
@click.argument("platform", type=str)
@click.argument("instrument_type", type=str)
@click.option("--format",
              type=click.Choice(ALLOWED_FORMATS),
              default=pyaurorax.search.FORMAT_BASIC_INFO,
              help="Amount of data about the data source to retrieve")
@click.option("--include-stats", is_flag=True, help="Include additional information about data sources")
@click.pass_obj
def get(config, program, platform, instrument_type, format, include_stats):
    """
    Get a single data source record

    \b
    PROGRAM           the program value
    PLATFORM          the platform value
    INSTRUMENT_TYPE   the instrument type value
    """
    # get the data source
    try:
        ds = config.aurorax.search.sources.get(program=program,
                                               platform=platform,
                                               instrument_type=instrument_type,
                                               format=format,
                                               include_stats=include_stats)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    __print_single_data_source(ds, format, include_stats)


@sources_group.command("get_using_identifier", short_help="Get a single data source (using an identifier)")
@click.argument("identifier", type=int)
@click.option("--format",
              type=click.Choice(ALLOWED_FORMATS),
              default=pyaurorax.search.FORMAT_BASIC_INFO,
              help="Amount of data about the data source to retrieve")
@click.option("--include-stats", is_flag=True, help="Include additional information about data sources")
@click.pass_obj
def get_using_identifier(config, identifier, format, include_stats):
    """
    Get a single data source record using an identifier

    \b
    IDENTIFIER     the identifier of the data source
    """
    # get data source
    try:
        ds = config.aurorax.search.sources.get_using_identifier(identifier, format=format, include_stats=include_stats)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    __print_single_data_source(ds, format, include_stats=include_stats)


@sources_group.command("add", short_help="Add a data source")
@click.argument("program", type=str)
@click.argument("platform", type=str)
@click.argument("instrument_type", type=str)
@click.argument("source_type", type=click.Choice(SUPPORTED_SOURCE_TYPES))
@click.argument("display_name", type=str)
@click.option("--identifier", type=int, help="Custom identifier to use")
@click.pass_obj
def add(config, program, platform, instrument_type, source_type, display_name, identifier):  # pragma: nocover
    """
    Add a data source

    \b
    PROGRAM           the program to set
    PLATFORM          the platform to set
    INSTRUMENT_TYPE   the instrument type to set
    SOURCE_TYPE       the source type to set
    DISPLAY_NAME      the display name to set
    """
    try:
        new_ds = config.aurorax.search.sources.DataSource(identifier=identifier,
                                                          program=program,
                                                          platform=platform,
                                                          instrument_type=instrument_type,
                                                          source_type=source_type,
                                                          display_name=display_name)
        added_ds = config.aurorax.search.sources.add(new_ds)
        click.echo("Created data source successfully\n")
        click.echo(added_ds)
    except pyaurorax.AuroraXError as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)


@sources_group.command("update", short_help="Update a data source")
@click.argument("identifier", type=str)
@click.option("--program", type=str, help="New program value")
@click.option("--platform", type=str, help="New platform value")
@click.option("--instrument-type", type=str, help="New instrument type value")
@click.option("--source-type", type=click.Choice(SUPPORTED_SOURCE_TYPES), help="New source type value")
@click.option("--display-name", type=str, help="New display name value")
@click.pass_obj
def update(config, identifier, program, platform, instrument_type, source_type, display_name):  # pragma: nocover
    """
    Update a data source

    \b
    IDENTIFIER     the identifier of the data source
    """
    try:
        ds = config.aurorax.search.sources.update_partial(identifier,
                                                          program=program,
                                                          platform=platform,
                                                          instrument_type=instrument_type,
                                                          source_type=source_type,
                                                          display_name=display_name)
        click.echo("Updated data source successfully\n")
        click.echo(ds)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)


@sources_group.command("delete", short_help="Delete a data source")
@click.argument("identifier", type=int)
@click.pass_obj
def delete(config, identifier):  # pragma: nocover
    """
    Delete a data source

    \b
    IDENTIFIER     the identifier of the data source
    """
    try:
        config.aurorax.search.sources.delete(identifier)
        click.echo("Successfully deleted data source #%d" % (identifier))
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
AuroraX data source record
"""

from typing import List, Dict, Optional
from .data_source_stats import DataSourceStatistics

# basic format
FORMAT_BASIC_INFO: str = "basic_info"
"""
Data sources are returned with basic information: identifier,
program, platform, instrument type, source type, and display name
"""

# basic, but with metadata as well
FORMAT_BASIC_INFO_WITH_METADATA: str = "with_metadata"
"""
Data sources are returned with basic information, plus the metadata
"""

# minimal, only the identifier
FORMAT_IDENTIFIER_ONLY: str = "identifier_only"
"""
Data sources are returned with only the identifier
"""

# full record, everything
FORMAT_FULL_RECORD: str = "full_record"
"""
Data sources are returned with all information.
"""

# default
FORMAT_DEFAULT: str = FORMAT_BASIC_INFO
"""
Default data source format (basic info)
"""

# ground source type
SOURCE_TYPE_GROUND: str = "ground"
"""
Data source 'source_type' category for a ground instrument
"""

# low-earth orbiting spacecraft source type
SOURCE_TYPE_LEO: str = "leo"
"""
Data source 'source_type' category for a low-earth orbiting satellite
"""

# highly-elliptical orbiting spacecraft source type
SOURCE_TYPE_HEO: str = "heo"
"""
Data source 'source_type' category for a highly-elliptical orbiting satellite
"""

# lunar orbiting spacecraft source type
SOURCE_TYPE_LUNAR: str = "lunar"
"""
Data source 'source_type' category for a lunar orbiting satellite
"""

# event list source type
SOURCE_TYPE_EVENT_LIST: str = "event_list"
"""
Data source 'source_type' category for a specially-curated event list
"""

# not applicable source type
SOURCE_TYPE_NOT_APPLICABLE: str = "not_applicable"
"""
Data source 'source_type' category for a specially-curated event list
"""


class DataSource:
    """
    AuroraX data source record

    Attributes:
        identifier (int): 
            The unique AuroraX data source identifier

        program (str): 
            The program for this data source

        platform (str): 
            The platform for this data source

        instrument_type (str): 
            The instrument type for this data source

        source_type (str): 
            The data source type for this data source. Options are in the 
            pyaurorax.search.sources module, or at the top level using the
            pyaurorax.search.SOURCE_TYPE_* variables.

        display_name (str): 
            The display name for this data source

        metadata (Dict): 
            Metadata for this data source (arbitrary keys and values)

        owner (str): 
            The owner's email address of this data source

        maintainers (List[str]): 
            The email addresses of AuroraX accounts that can alter this data source and 
            its associated records

        ephemeris_metadata_schema (Dict): 
            A list of dictionaries capturing the metadata keys and values that can appear 
            in ephemeris records associated with this data source

        data_product_metadata_schema (Dict): 
            A list of dictionaries capturing the metadata keys and values that can appear 
            in data product records associated with this data source

        stats (DataSourceStatistics): 
            Data source statistics information
            
        format (str): 
            The format used when printing the data source, defaults to `full_record`. Other 
            options are in the pyaurorax.search.sources module, or at the top level using 
            the pyaurorax.search.FORMAT_* variables.
    """

    def __init__(self,
                 identifier: Optional[int] = None,
                 program: Optional[str] = None,
                 platform: Optional[str] = None,
                 instrument_type: Optional[str] = None,
                 source_type: Optional[str] = None,
                 display_name: Optional[str] = None,
                 metadata: Optional[Dict] = None,
                 owner: Optional[str] = None,
                 maintainers: Optional[List[str]] = None,
                 ephemeris_metadata_schema: Optional[List[Dict]] = None,
                 data_product_metadata_schema: Optional[List[Dict]] = None,
                 stats: Optional[DataSourceStatistics] = None,
                 format: str = FORMAT_FULL_RECORD):
        self.identifier = identifier
        self.program = program
        self.platform = platform
        self.instrument_type = instrument_type
        self.source_type = source_type
        self.display_name = display_name
        self.metadata = metadata
        self.owner = owner
        self.maintainers = maintainers
        self.ephemeris_metadata_schema = ephemeris_metadata_schema
        self.data_product_metadata_schema = data_product_metadata_schema
        self.stats = stats
        self.format = format

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "DataSource(identifier=%s, program='%s', platform='%s', instrument_type='%s', source_type='%s', display_name='%s', ...)" % (
            self.identifier,
            self.program,
            self.platform,
            self.instrument_type,
            self.source_type,
            self.display_name,
        )

    def pretty_print(self):
        """
        A special print output for this class.
        """
        max_len = 80
        print("DataSource:")
        print("  %-30s: %d" % ("identifier", self.identifier))
        print("  %-30s: %s" % ("program", self.program))
        print("  %-30s: %s" % ("platform", self.platform))
        print("  %-30s: %s" % ("instrument_type", self.instrument_type))
        print("  %-30s: %s" % ("source_type", self.source_type))
        print("  %-30s: %s" % ("display_name", self.display_name))
        print("  %-30s: %s" % ("metadata", self.metadata))
        print("  %-30s: %s" % ("owner", self.owner))
        print("  %-30s: %s" % ("maintainers", self.maintainers))
        if (self.ephemeris_metadata_schema is not None and len(str(self.ephemeris_metadata_schema)) > max_len):
            ephemeris_metadata_schema_str = "%s..." % (str(self.ephemeris_metadata_schema)[0:max_len])
        else:
            ephemeris_metadata_schema_str = self.ephemeris_metadata_schema
        print("  %-30s: %s" % ("ephemeris_metadata_schema", ephemeris_metadata_schema_str))
        if (self.data_product_metadata_schema is not None and len(str(self.data_product_metadata_schema)) > max_len):
            data_product_metadata_schema_str = "%s..." % (str(self.data_product_metadata_schema)[0:max_len])
        else:
            data_product_metadata_schema_str = self.data_product_metadata_schema
        print("  %-30s: %s" % ("data_product_metadata_schema", data_product_metadata_schema_str))
        if (self.stats is not None and len(str(self.stats)) > max_len):  # pragma: nocover
            stats_str = "%s..." % (str(self.stats)[0:max_len])
        else:
            stats_str = self.stats
        print("  %-30s: %s" % ("stats", stats_str))
        print("  %-30s: %s" % ("format", self.format))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for using conjunction searches with Swarm-Aurora
"""

import webbrowser
import json
from ...api import AuroraXAPIRequest


def get_url(search_obj):
    return "https://swarm-aurora.com/conjunctionFinder?aurorax_request_id=%s" % (search_obj.request_id)


def open_in_browser(search_obj, browser):
    url = get_url(search_obj)
    try:
        w = webbrowser.get(using=browser)
        w.open_new_tab(url)
    except Exception as e:
        if ("could not locate runnable browser" in str(e)):
            raise ValueError(("Error: selected browser '%s' not found, please try another. For the list of options, refer to "
                              "https://docs.python.org/3/library/webbrowser.html#webbrowser.get") % (browser)) from e


def create_custom_import_file(aurorax_obj, search_obj, filename, return_dict):
    # make request
    url = "https://swarm-aurora.com/conjunctionFinder/generate_custom_import_json?aurorax_request_id=%s" % (search_obj.request_id)
    req = AuroraXAPIRequest(aurorax_obj, method="get", url=url, body=search_obj.query)
    res = req.execute()

    # return the contents as a dict if requested
    if (return_dict is True):
        return res.data

    # set default filename
    if (filename is None):
        filename = "swarmaurora_custom_import_%s.json" % (search_obj.request_id)

    # save data to file
    with open(filename, 'w', encoding="utf-8") as fp:
        json.dump(res.data, fp, indent=4)

    # return
    return filename



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Generate keograms.
"""

import datetime
import numpy as np
from typing import Optional, List, Literal, Tuple, Union
from ..classes.keogram import Keogram
from ...data.ucalgary import Skymap
from ._create import create as func_create
from ._create_custom import create_custom as func_create_custom

__all__ = ["KeogramManager"]


class KeogramManager:
    """
    The KeogramManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self):
        pass

    def create(self,
               images: np.ndarray,
               timestamp: List[datetime.datetime],
               axis: int = 0,
               spectra: bool = False,
               wavelength: Optional[np.ndarray] = None,
               spect_emission: Literal["green", "red", "blue", "hbeta"] = "green",
               spect_band: Optional[Tuple[float, float]] = None,
               spect_band_bg: Optional[Tuple[float, float]] = None) -> Keogram:
        """
        Create a keogram from a set of images.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also 
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images]. If it is not, then be sure to specify the `axis` parameter
                accordingly.

            timestamp (List[datetime.datetime]): 
                A list of timestamps corresponding to each image.

            axis (int): 
                The axis to extract the keogram slice from. Default is `0`, meaning the rows (or Y) axis.

            spectra (bool): 
                Make a keogram out of spectrograph data, for a specific emission. Defaults to False (ASI data).

            wavelength (numpy.ndarray): 
                The wavelength array corresponding to spectrograph data. If spectra=True, this parameter
                must be supplied.

            spect_emission (str): 
                The emission (green, red, blue, hbeta) to prepare from spectrograph data. Default is 
                'green' (557.7 nm emission).

            spect_band (Tuple[float]): 
                Manual selection of the wavelength region to integrate for obtaining emissions. Use this
                to prepare emissions that are not available in spect_emission.

            spect_band_bg (Tuple[float]): 
                Manual selection of the wavelength region to subtract from integration for manually
                chosen emissions, via the spect_band argument.

        Returns:
            A `pyaurorax.tools.Keogram` object.

        Raises:
            ValueError: issues encountered with supplied parameters
        """
        return func_create(images, timestamp, axis, spectra, wavelength, spect_emission, spect_band, spect_band_bg)

    def create_custom(
        self,
        images: np.ndarray,
        timestamp: List[datetime.datetime],
        coordinate_system: Literal["ccd", "geo", "mag"],
        width: int,
        x_locs: Union[List[Union[float, int]], np.ndarray],
        y_locs: Union[List[Union[float, int]], np.ndarray],
        preview: bool = False,
        skymap: Optional[Skymap] = None,
        altitude_km: Optional[Union[float, int]] = None,
        metric: Literal["mean", "median", "sum"] = "median",
    ) -> Keogram:
        """
        Create a keogram, from a custom slice of a set of images. The slice used is defined by a set of points, 
        in CCD, geographic, or geomagnetic coordinates, within the bounds of the image data. Keogram is created
        from the bottom up, meaning the first point will correspond to the bottom of the keogram data.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also 
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images]. If it is not, then be sure to specify the `axis` parameter
                accordingly.

            timestamp (List[datetime.datetime]): 
                A list of timestamps corresponding to each image.
            
            coordinate_system (str): 
                The coordinate system in which input points are defined. Valid options are "ccd", "geo", or "mag".
            
            width (int): 
                Width of the desired keogram slice, in CCD pixel units.

            x_locs (Sequence[float, int]): 
                Sequence of points giving the x-coordinates that define a path through the image data, from
                which to build the keogram.

            y_locs (Sequence[float, int]): 
                Sequence of points giving the y-coordinates that define a path through the image data, from
                which to build the keogram.

            preview (Optional[bool]): 
                When True, the first frame in images will be displayed, with the keogram slice plotted.

            skymap (Skymap): 
                The skymap to use in georeferencing when working in geographic or magnetic coordinates.

            altitude_km (float, int): 
                The altitude of the image data, in km, to use in georeferencing when working in geographic
                or magnetic coordinates.

            metric (str): 
                The metric used to compute values for each keogram pixel. Valid options are "median", "mean",
                and "sum". Defaults to "median".
            
        Returns:
            A `pyaurorax.tools.Keogram` object.

        Raises:
            ValueError: issues encountered with supplied parameters
        """
        return func_create_custom(images, timestamp, coordinate_system, width, x_locs, y_locs, preview, skymap, altitude_km, metric)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for an AuroraX API response
"""

from typing import Any
from http.client import responses


class AuroraXAPIResponse:
    """
    Class definition for an AuroraX API response

    Attributes:
        request (Any): 
            The request object

        data (Any): 
            The data received as part of the request

        status_code (int): 
            The HTTP status code received when making the request
    """

    def __init__(self, request: Any, data: Any, status_code: int):
        self.request = request
        self.data = data
        self.status_code = status_code

    def __str__(self) -> str:
        return self.__repr__()  # pragma: nocover-ok

    def __repr__(self) -> str:
        return f"AuroraXAPIResponse [{self.status_code}] ({responses[self.status_code]})"  # pragma: nocover-ok



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Prepare data and create mosaics.
"""

import datetime
import cartopy.crs
from typing import Union, Optional, List, Dict, Tuple, Literal
from ..._util import show_warning
from ...data.ucalgary import Data, Skymap
from ..classes.mosaic import MosaicData, MosaicSkymap, Mosaic
from ._prep_skymaps import prep_skymaps as func_prep_skymaps
from ._prep_images import prep_images as func_prep_images
from ._create import create as func_create

__all__ = ["MosaicManager"]


class MosaicManager:
    """
    The MosaicManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

    def create(
            self,
            prepped_data: Union[MosaicData, List[MosaicData]],
            prepped_skymap: Union[MosaicSkymap, List[MosaicSkymap]],
            timestamp: datetime.datetime,
            cartopy_projection: cartopy.crs.Projection,
            min_elevation: Union[int, List[int]] = 5,
            cmap: Optional[Union[str, List[str]]] = None,
            spect_cmap: Optional[Union[str, List[str]]] = None,
            image_intensity_scales: Optional[Union[List, Dict]] = None,
            spect_intensity_scales: Optional[Tuple[int, int]] = None,
            colormap: Optional[Union[str, List[str]]] = None,  # deprecated in v1.10.0
            spect_colormap: Optional[Union[str, List[str]]] = None,  # deprecated in v1.10.0
    ) -> Mosaic:
        """
        Create a mosaic object.

        Args:
            prepped_data (pyaurorax.tools.MosaicData): 
                The prepared mosaic data. Generated from a prior `prep_images()` function call.

            prepped_skymap (pyaurorax.tools.MosaicSkymap): 
                The prepared skymap data. Generated from a prior `prep_skymaps()` function call.

            timestamp (datetime.datetime): 
                The timestamp to generate a mosaic for. Must be within the range of timestamps
                for which image data was prepped and provided.

            cartopy_projection (cartopy.crs.Projection): 
                The cartopy projection to use when creating the mosaic.

            min_elevation (int): 
                The minimum elevation cutoff when projecting images on the map, in degrees. Default is `5`.

            cmap (str): 
                The matplotlib colormap to use for the rendered image data. Default is `gray`.

                Commonly used colormaps are:

                - REGO: `gist_heat`
                - THEMIS ASI: `gray`
                - TREx Blue: `Blues_r`
                - TREx NIR: `gray`
                - TREx RGB: `None`

                A list of all available colormaps can be found on the 
                [matplotlib documentation](https://matplotlib.org/stable/gallery/color/colormap_reference.html).

            colormap (str): 
                Deprecated as of v1.10.0. Use 'cmap' instead in the exact same way.

            spect_cmap (str): 
                The matplotlib colormap to use for the colorbar if working with spectrograph
                data. Default is `gnuplot`.

            image_intensity_scales (List or Dict): 
                Ranges for scaling images. Either a a list with 2 elements which will scale all sites with 
                the same range, or as a dictionary which can be used for scaling each site differently. 

                Example of uniform scaling across all sites: 
                `image_intensity_scales = [2000, 8000]`

                Example of scaling each site differently:
                `image_intensity_scales = {"fsmi": [1000, 10000], "gill": [2000, 8000]}`

            spect_intensity_scales (Tuple[int]): 
                Min and max values, in Rayleighs, to scale ALL spectrograph data.

            spect_colormap (str): 
                The name of a matplotlib colormap to use for plotting spectrograph data.

        Returns:
            The generated `pyaurorax.tools.Mosaic` object.

        Raises:
            ValueError: issues encountered with supplied parameters
            pyaurorax.exceptions.AuroraXError: general issue encountered
        """
        # handle deprecation warnings
        use_colormap = False
        use_spect_colormap = False
        if (colormap is not None):
            show_warning("The parameter 'colormap' was deprecated in v1.10.0. Please use 'cmap' instead (usage is identical).", stacklevel=1)
            use_colormap = True
        if (spect_colormap is not None):
            show_warning("The parameter 'spect_colormap' was deprecated in v1.10.0. Please use 'spect_cmap' instead (usage is identical).",
                         stacklevel=1)
            use_spect_colormap = True

        # run function
        if (use_colormap is False and use_spect_colormap is False):
            # normal
            return func_create(prepped_data, prepped_skymap, timestamp, cartopy_projection, min_elevation, cmap, spect_cmap, image_intensity_scales,
                               spect_intensity_scales)
        elif (use_colormap is True and use_spect_colormap is True):
            # both deprecated parameters were supplied
            return func_create(prepped_data, prepped_skymap, timestamp, cartopy_projection, min_elevation, colormap, spect_colormap,
                               image_intensity_scales, spect_intensity_scales)
        elif (use_colormap is True and use_spect_colormap is False):
            # one deprecated parameter was supplied (colormap)
            return func_create(prepped_data, prepped_skymap, timestamp, cartopy_projection, min_elevation, colormap, spect_cmap,
                               image_intensity_scales, spect_intensity_scales)
        else:
            # one deprecated parameter was supplied (spect_colormap)
            return func_create(prepped_data, prepped_skymap, timestamp, cartopy_projection, min_elevation, cmap, spect_colormap,
                               image_intensity_scales, spect_intensity_scales)

    def prep_images(self,
                    image_list: List[Data],
                    data_attribute: Literal["data", "calibrated_data"] = "data",
                    spect_emission: Literal["green", "red", "blue", "hbeta"] = "green",
                    spect_band: Optional[Tuple[float, float]] = None,
                    spect_band_bg: Optional[Tuple[float, float]] = None) -> MosaicData:
        """
        Prepare the image data for use in a mosaic.

        Args:
            image_list (List[pyaurorax.data.ucalgary.Data]): 
                List of image data. Each element of the list is the data for each site.
            
            data_attribute (str): 
                The data attribute to use when prepping the images. Either `data` or `calibrated_data`. 
                Default is `data`.

            spect_emission (str): 
                The emission (green, red, blue, hbeta) to prepare from spectrograph data. Default is 
                'green' (557.7 nm emission).

            spect_band (Tuple[float]): 
                Manual selection of the wavelength region to integrate for obtaining emissions. Use this
                to prepare emissions that are not available in spect_emission.

            spect_band_bg (Tuple[float]): 
                Manual selection of the wavelength region to subtract from integration for manually
                chosen emissions, via the spect_band argument.

        Returns:
            The prepared data, as a `pyaurorax.tools.MosaicData` object.

        Raises:
            ValueError: issues encountered with supplied parameters
        """
        return func_prep_images(image_list, data_attribute, spect_emission, spect_band, spect_band_bg)

    def prep_skymaps(self,
                     skymaps: List[Skymap],
                     height_km: int,
                     site_uid_order: Optional[List[str]] = None,
                     progress_bar_disable: bool = False,
                     n_parallel: int = 1) -> MosaicSkymap:
        """
        Prepare skymap data for use by the mosaic routine. This is not time-dependent, so it 
        would only need to be done once.

        Allows for plotting multiple images on a map, masking the boundaries between 
        images by elevation angle.

        Args:
            skymaps (List[pyaurorax.data.ucalgary.Skymap]): 
                The skymaps to prep.
            
            height_km (int): 
                The altitude to utilize, in kilometers.
            
            site_uid_order (List[str]): 
                The site list order. The order of this list is not important for plotting, but must be
                consistent with the order of the `skymaps` parameter.
            
            progress_bar_disable (bool): 
                Disable the progress bar. Defaults to `False`.

            n_parallel (int): 
                Number of skymaps to prepare in parallel using multiprocessing. Default is `1`. 

        Returns:
            The prepared skymap data as a `pyaurorax.tools.MosaicSkymap` object.
            
        Raises:
            ValueError: issues encountered with supplied parameters
        """
        return func_prep_skymaps(self.__aurorax_obj, skymaps, height_km, site_uid_order, progress_bar_disable, n_parallel)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for AuroraX search engine location data
"""

from typing import Optional, Dict


class Location:
    """
    Representation for an AuroraX search engine location. This data can be in geodetic 
    coordinates, GSM coordinates, or geodetic northern/southern B-trace magnetic footprints.

    Latitude and longitude values are in decimal degrees format, ranging from -90 to 90
    for latitude and -180 to 180 for longitude.

    Note that latitude and longitude must both be numbers, or both be None.

    Attributes:
        lat (float): latitude value
        lon (float): longitude value
    
    Raises:
        ValueError: if both latitude and longitude are not real numbers, or not both None.
    """

    def __init__(self, lat: Optional[float] = None, lon: Optional[float] = None):
        if (lat is None and lon is not None) or (lat is not None and lon is None):
            # one of them is None, not allowed
            raise ValueError("Latitude and longitude must both be numbers, or both be None")
        self.__lat = lat
        self.__lon = lon

    @property
    def lat(self):
        return self.__lat

    @lat.setter
    def lat(self, value: float):
        if (self.__lon is None and value is not None) or (self.__lon is not None and value is None):
            # one of them is None, not allowed
            raise ValueError("Latitude and longitude must both be numbers, or both be None")
        self.__lat = value

    @property
    def lon(self):
        return self.__lon

    @lon.setter
    def lon(self, value: float):
        if (self.__lat is None and value is not None) or (self.__lat is not None and value is None):
            # one of them is None, not allowed
            raise ValueError("Latitude and longitude must both be numbers, or both be None")
        self.__lon = value

    def to_json_serializable(self) -> Dict:
        """
        Convert object to a JSON-serializable object (ie. translate
        datetime objects to strings)

        Returns:
            a dictionary object that is JSON-serializable
        """
        return {"lat": self.lat, "lon": self.lon}

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "%s(lat=%s, lon=%s)" % (self.__class__.__name__, str(self.lat), str(self.lon))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for performing data product searches
"""

import datetime
import humanize
import numpy as np
from .classes.data_product import DataProductData
from .classes.search import DataProductSearch
from ..api import AuroraXAPIRequest
from ..sources.classes.data_source import FORMAT_DEFAULT
from ..sources._sources import get_using_identifier
from ...exceptions import (
    AuroraXError,
    AuroraXAPIError,
    AuroraXSearchError,
    AuroraXUploadError,
)

__STANDARD_POLLING_SLEEP_TIME: float = 1.0  # Polling sleep time when waiting for data (after the initial sleep time), in seconds


def __validate_data_source(aurorax_obj, identifier, records):
    # get data source
    try:
        ds = get_using_identifier(aurorax_obj, identifier, FORMAT_DEFAULT, False)
    except AuroraXAPIError as e:  # pragma: nocover-ok
        if ("no data source record found" in str(e).lower()):
            raise AuroraXAPIError("Data source with identifier %d could not be found" % (identifier)) from e
        else:
            raise e

    # process each record to make sure the program/platform/instrument_type matches
    # the identifier found for the data source
    for record in records:
        if not (record.data_source.program == ds.program and record.data_source.platform == ds.platform
                and record.data_source.instrument_type == ds.instrument_type):
            return record

    # found no bad records
    return None


def search(aurorax_obj, start, end, programs, platforms, instrument_types, data_product_types, metadata_filters, metadata_filters_logical_operator,
           response_format, poll_interval, return_immediately, verbose):
    # create a search object
    s = DataProductSearch(aurorax_obj,
                          start,
                          end,
                          programs=programs,
                          platforms=platforms,
                          instrument_types=instrument_types,
                          data_product_types=data_product_types,
                          metadata_filters=metadata_filters,
                          metadata_filters_logical_operator=metadata_filters_logical_operator,
                          response_format=response_format)
    if (verbose is True):
        print("[%s] Search object created" % (datetime.datetime.now()))

    # execute the search
    s.execute()
    if (verbose is True):
        print("[%s] Request submitted" % (datetime.datetime.now()))
        print("[%s] Request ID: %s" % (datetime.datetime.now(), s.request_id))
        print("[%s] Request details available at: %s" % (datetime.datetime.now(), s.request_url))

    # return immediately if we wanted to
    if (return_immediately is True):
        return s

    # wait for data
    if (verbose is True):
        print("[%s] Waiting for data ..." % (datetime.datetime.now()))
    s.wait(poll_interval=poll_interval, verbose=verbose)

    # check if error condition encountered
    if (s.status["search_result"]["error_condition"] is True):  # pragma: nocover-ok
        # error encountered
        raise AuroraXSearchError(s.logs[-1]["summary"])

    # get the data
    if (verbose is True):
        print("[%s] Retrieving data ..." % (datetime.datetime.now()))
    s.get_data()

    # return response with the data
    if (verbose is True):
        print("[%s] Retrieved %s of data containing %d records" % (
            datetime.datetime.now(),
            humanize.filesize.naturalsize(s.status["search_result"]["file_size"]),  # type: ignore
            s.status["search_result"]["result_count"],
        ))
    return s


def upload(aurorax_obj, identifier, all_records, validate_source, chunk_size):
    # validate record sources if the flag is set
    if (validate_source is True):
        validation_error = __validate_data_source(aurorax_obj, identifier, all_records)
        if (validation_error is not None):
            raise AuroraXError("Unable to validate data source found in record: {}".format(validation_error))

    # translate each data product record to a request-friendly
    # dict (ie. convert datetimes to strings, etc.)
    for i, _ in enumerate(all_records):
        if (isinstance(all_records[i], DataProductData) is True):
            all_records[i] = all_records[i].to_json_serializable()

    # chunk up the records
    #
    # NOTE: we do this so that upload requests that are large are able to
    # more consistently succeed. If a call is to upload 10,000 records, we
    # under-the-hood chunk it up into N-record calls.
    if (len(all_records) > chunk_size):  # pragma: nocover
        chunked_indexes = np.array_split(np.array(range(0, len(all_records)), dtype=np.int32), int(np.floor(len(all_records) / chunk_size)))
    else:
        chunked_indexes = [np.arange(0, len(all_records), dtype=np.int32)]
    for idxs in chunked_indexes:
        records = []
        for idx in idxs:
            records.append(all_records[idx])

        # make request
        url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_UPLOAD.format(identifier))
        req = AuroraXAPIRequest(aurorax_obj, method="post", url=url, body=records, null_response=True)
        res = req.execute()

        # evaluate response
        if (res.status_code == 400):
            if isinstance(res.data, list):  # pragma: nocover-ok
                raise AuroraXUploadError("%s - %s" % (res.status_code, res.data[0]["message"]))
            raise AuroraXUploadError("%s - %s" % (res.data["error_code"], res.data["error_message"]))  # pragma: nocover

    # return
    return 0


def delete_urls(aurorax_obj, data_source, urls):
    # check to make sure the identifier, program, platform, and instrument type are all set in the data source
    if not all([data_source.identifier, data_source.program, data_source.platform, data_source.instrument_type]):  # pragma: nocover-ok
        raise AuroraXError("One or more required data source parameters are missing, delete operation aborted")

    # do request
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_UPLOAD.format(data_source.identifier))
    params = {
        "program": data_source.program,
        "platform": data_source.platform,
        "instrument_type": data_source.instrument_type,
        "urls": urls,
    }
    delete_req = AuroraXAPIRequest(aurorax_obj, method="delete", url=url, body=params, null_response=True)
    res = delete_req.execute()

    # evaluate response
    if (res.status_code == 400):  # pragma: nocover-ok
        raise AuroraXAPIError("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 0


def delete(aurorax_obj, data_source, start, end, data_product_types):
    # check to make sure the identifier, program, platform, and instrument type are all set in the data source
    if not all([data_source.identifier, data_source.program, data_source.platform, data_source.instrument_type]):  # pragma: nocover-ok
        raise AuroraXError("One or more required data source parameters are missing, delete operation aborted")

    # do request to get all data products between start and end datetimes
    try:
        s = search(aurorax_obj,
                   start,
                   end,
                   programs=[data_source.program],
                   platforms=[data_source.platform],
                   instrument_types=[data_source.instrument_type],
                   data_product_types=[] if not data_product_types else data_product_types,
                   metadata_filters=None,
                   metadata_filters_logical_operator=None,
                   response_format=None,
                   poll_interval=__STANDARD_POLLING_SLEEP_TIME,
                   return_immediately=False,
                   verbose=False)
    except Exception as e:  # pragma: nocover-ok
        raise AuroraXError(e) from e

    # collect URLs from search result
    urls = []
    for dp in s.data:
        urls.append(dp.url)

    # do delete request
    return delete_urls(aurorax_obj, data_source, urls)


def describe(aurorax_obj, search_obj, query_dict):
    # set query
    if (search_obj is not None):
        query = search_obj.query
    elif (query_dict is not None):
        query = query_dict
    else:
        raise AuroraXError("One of 'search_obj' or 'query_dict' must be supplied")

    # make request
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DESCRIBE_DATA_PRODUCTS_QUERY)
    req = AuroraXAPIRequest(aurorax_obj, method="post", url=url, body=query)
    res = req.execute()

    # return
    return res.data


def get_request_url(aurorax_obj, request_id):
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_REQUEST.format(request_id))
    return url



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Extract various metrics from a given bounding box.
"""

import datetime
import numpy as np
from typing import Sequence, Union, Literal, Optional
from ....data.ucalgary import Skymap
from ._azimuth import azimuth as func_azimuth
from ._ccd import ccd as func_ccd
from ._elevation import elevation as func_elevation
from ._geo import geo as func_geo
from ._mag import mag as func_mag

__all__ = ["ExtractMetricManager"]


class ExtractMetricManager:
    """
    The ExtractMetricManager object is initialized within every PyAuroraX object. It acts as a way to 
    access the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

    def azimuth(self,
                images: np.ndarray,
                skymap: Skymap,
                azimuth_bounds: Sequence[Union[int, float]],
                metric: Literal["mean", "median", "sum"] = "median",
                n_channels: Optional[int] = None,
                show_preview: bool = False) -> np.ndarray:
        """
        Compute a metric of image data within an azimuthal boundary.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images].
            
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the image data.

            azimuth_bounds (Sequence[int, float]): 
                A 2-element sequence specifying the azimuthal bounds from which to extract the metric. 
                Anticipated order is [az_min, az_max].

            metric (str): 
                The name of the metric that is to be computed for the bounded area. Valid metrics are `mean`,
                `median`, `sum`. Default is `median`.

            n_channels (int): 
                By default, function will assume the type of data passed as input - this argument can be used
                to manually specify the number of channels contained in image data.

            show_preview (bool): 
                Plot a preview of the bounded area.

        Returns:
            A numpy.ndarray containing the metrics computed within azimuth range, for all image frames.

        Raises:
            ValueError: issue encountered with value supplied in parameter
        """
        return func_azimuth(self.__aurorax_obj, images, skymap, azimuth_bounds, metric, n_channels, show_preview)

    def ccd(self,
            images: np.ndarray,
            ccd_bounds: Sequence[int],
            metric: Literal["mean", "median", "sum"] = "median",
            n_channels: Optional[int] = None,
            show_preview: bool = False) -> np.ndarray:
        """
        Compute a metric of image data within a CCD boundary.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images].
            
            ccd_bounds (List[int]): 
                A 4-element sequence specifying the (inclusive) CCD bounds from which to extract the metric. 
                Anticipated order is [x0, x1, y0, y1].

            metric (str): 
                The name of the metric that is to be computed for the bounded area. Valid metrics are `mean`,
                `median`, `sum`. Defaults to `median`.

            n_channels (int): 
                By default, function will assume the type of data passed as input - this argument can be used
                to manually specify the number of channels contained in image data.
            
            show_preview (bool): 
                Plot a preview of the bounded area.

        Returns:
            A numpy.ndarray containing the metrics computed within CCD bounds, for all image frames.

        Raises:
            ValueError: issue encountered with value supplied in parameter
        """
        return func_ccd(self.__aurorax_obj, images, ccd_bounds, metric, n_channels, show_preview)

    def elevation(self,
                  images: np.ndarray,
                  skymap: Skymap,
                  elevation_bounds: Sequence[Union[int, float]],
                  metric: Literal["mean", "median", "sum"] = "median",
                  n_channels: Optional[int] = None,
                  show_preview: bool = False) -> np.ndarray:
        """
        Compute a metric of image data within an elevation boundary.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images].
            
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the image data.

            elevation_bounds (Sequence): 
                A 2-element sequence specifying the elevation bounds from which to extract the metric. 
                Anticipated order is [el_min, el_max].

            metric (str): 
                The name of the metric that is to be computed for the bounded area. Valid metrics are `mean`,
                `median`, `sum`. Default is `median`.

            n_channels (int): 
                By default, function will assume the type of data passed as input - this argument can be used
                to manually specify the number of channels contained in image data.
            
            show_preview (bool): 
                Plot a preview of the bounded area.


        Returns:
            A numpy.ndarray containing the metrics computed within elevation range, for all image frames.

        Raises:
            ValueError: issue encountered with value supplied in parameter
        """
        return func_elevation(self.__aurorax_obj, images, skymap, elevation_bounds, metric, n_channels, show_preview)

    def geo(self,
            images: np.ndarray,
            skymap: Skymap,
            altitude_km: Union[int, float],
            lonlat_bounds: Sequence[Union[int, float]],
            metric: Literal["mean", "median", "sum"] = "median",
            n_channels: Optional[int] = None,
            show_preview: bool = False) -> np.ndarray:
        """
        Compute a metric of image data within a geographic lat/lon boundary.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images].
            
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the image data.
            
            altitude_km (int or float): 
                The altitude of the image data in kilometers.

            lonlat_bounds (Sequence): 
                A 4-element sequence specifying the lat/lon bounds from which to extract the metric. 
                Anticipated order is [lon_0, lon_1, lat_0, lat_1].

            metric (str): 
                The name of the metric that is to be computed for the bounded area. Valid metrics are `mean`,
                `median`, `sum`. Default is `median`.

            n_channels (int): 
                By default, function will assume the type of data passed as input - this argument can be used
                to manually specify the number of channels contained in image data.
            
            show_preview (bool): 
                Plot a preview of the bounded area.

        Returns:
            A numpy.ndarray containing the metrics computed within elevation range, for all image frames.

        Raises:
            ValueError: issue encountered with value supplied in parameter
        """
        return func_geo(self.__aurorax_obj, images, skymap, altitude_km, lonlat_bounds, metric, n_channels, show_preview)

    def mag(self,
            images: np.ndarray,
            timestamp: datetime.datetime,
            skymap: Skymap,
            altitude_km: Union[int, float],
            lonlat_bounds: Sequence[Union[int, float]],
            metric: Literal["mean", "median", "sum"] = "median",
            n_channels: Optional[int] = None,
            show_preview: bool = False) -> np.ndarray:
        """
        Compute a metric of image data within a magnetic lat/lon boundary.

        Args:
            images (numpy.ndarray): 
                A set of images. Normally this would come directly from a data `read` call, but can also
                be any arbitrary set of images. It is anticipated that the order of axes is [rows, cols, num_images]
                or [row, cols, channels, num_images].
            
            timestamp (List[datetime.datetime]): 
                A list of timestamps corresponding to each image.
            
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the image data.
            
            altitude_km (int or float): 
                The altitude of the image data in kilometers.

            lonlat_bounds (Sequence): 
                A 4-element sequence specifying the magnetic lat/lon bounds from which to extract the metric. 
                Anticipated order is [lon_0, lon_1, lat_0, lat_1].

            metric (str): 
                The name of the metric that is to be computed for the bounded area. Valid metrics are `mean`,
                `median`, `sum`. Default is `median`.

            n_channels (int): 
                By default, function will assume the type of data passed as input - this argument can be used
                to manually specify the number of channels contained in image data.

            show_preview (bool): 
                Plot a preview of the bounded area.

        Returns:
            A numpy.ndarray containing the metrics computed within elevation range, for all image frames.

        Raises:
            ValueError: issue encountered with value supplied in parameter
        """
        return func_mag(self.__aurorax_obj, images, timestamp, skymap, altitude_km, lonlat_bounds, metric, n_channels, show_preview)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt


def prep_grid_image(aurorax_obj, grid, fill_val, scale, cmap):
    # Determine number of channels in grid
    if len(grid.shape) == 2:
        n_channels = 1
    elif len(grid.shape) == 3:
        n_channels = 3
    else:
        raise ValueError("Routine currently only supports grid data with shape [rows, cols] or [rows, cols, channels].")

    # Convert grid to int type for plotting
    if np.issubdtype(grid.dtype, np.floating):
        grid = grid.astype(int)

    # Check that scale input is defined properly
    if scale is not None:
        # convert list to array
        if (isinstance(scale, tuple) or isinstance(scale, list)):
            scale = np.asarray(scale)
        elif (isinstance(scale, int) or isinstance(scale, float)):
            raise ValueError("Scale must be provided as a two-element vector, i.e. [scale_min, scale_max].")
        if len(scale) != 2:
            raise ValueError("Scale must be provided as a two-element vector, i.e. [scale_min, scale_max].")

    # Handle RGB data first as it is straightforward
    if n_channels == 3:
        # Replace fill values with nans
        grid = np.where(grid == fill_val, np.nan, grid)

        # Scale the image data
        if scale is None:
            scaled_grid = aurorax_obj.tools.scale_intensity(np.where(np.isnan(grid), 0, grid), top=255, memory_saver=False)
        else:
            scaled_grid = aurorax_obj.tools.scale_intensity(np.where(np.isnan(grid), 0, grid),
                                                            min=scale[0],
                                                            max=scale[1],
                                                            top=255,
                                                            memory_saver=False)

        # Add alpha channel
        alpha_channel = np.where(np.isnan(grid).any(axis=-1), 0, 255)
        rgba_grid = np.dstack((scaled_grid, alpha_channel))

        return rgba_grid.astype(int)

    # Handle case of single-channel data, where we build an RGBA image using a matplotlib colormap
    else:
        # Replace fill values with nans
        grid = np.where(grid == fill_val, np.nan, grid)

        # Scale the image data
        if scale is None:
            scaled_grid = aurorax_obj.tools.scale_intensity(np.where(np.isnan(grid), 0, grid), top=255, memory_saver=False)
        else:
            scaled_grid = aurorax_obj.tools.scale_intensity(np.where(np.isnan(grid), 0, grid),
                                                            min=scale[0],
                                                            max=scale[1],
                                                            top=255,
                                                            memory_saver=False)

        # Normalize array and apply colormap to obtain RGBA Array
        norm_scaled_grid = scaled_grid / 255.0
        colormap = plt.get_cmap(cmap)
        norm_scaled_grid = colormap(norm_scaled_grid)

        # 'De-normalize' array back to 8-bit range
        rgba_grid = (norm_scaled_grid * 255).astype(int)

        # Add alpha channel with max transparency for cells with no data
        alpha_channel = np.where(np.isnan(grid), 0, 255)
        rgba_grid[..., -1] = alpha_channel

        # return
        return rgba_grid



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
from tqdm.contrib.concurrent import process_map as tqdm_process_map
from concurrent.futures import ProcessPoolExecutor


def __process_frame(fname):  # pragma: nocover-ok
    return {
        "fname": fname,
        "img": cv2.imread(fname, cv2.IMREAD_UNCHANGED),
    }


def movie(aurorax_obj, input_filenames, output_filename, n_parallel, fps, progress_bar_disable):
    # read in all frames
    #
    # NOTE: we do with using multiprocessing, but we need to be very
    # careful to ensure that the frames are being added to the list
    # in order. This preserves the order of the file list that was
    # supplied.
    frame_list = []
    if (n_parallel == 1):
        # don't do anything special, just a basic loop
        if (progress_bar_disable is True):
            # no progress bar
            for fname in input_filenames:
                img = cv2.imread(fname)
                frame_list.append(img)
        else:
            # with progress bar
            for fname in aurorax_obj._tqdm(input_filenames, desc="Reading files: ", unit="files"):
                img = cv2.imread(fname)
                frame_list.append(img)
    else:
        # multiple workers, do it in a multiprocessing loop
        if (progress_bar_disable is True):
            with ProcessPoolExecutor(max_workers=n_parallel) as executor:
                for result in executor.map(__process_frame, input_filenames):
                    frame_list.append(result["img"])
        else:
            results = tqdm_process_map(
                __process_frame,
                input_filenames,
                max_workers=n_parallel,
                chunksize=1,
                desc="Reading files: ",
                unit="files",
                tqdm_class=aurorax_obj._tqdm,
            )
            for result in results:
                frame_list.append(result["img"])

    # get dimensions of images so we can initialize the video writer
    #
    # NOTE: we assume that all images are the same size. Perhaps we can
    # put a check in while we read all the images in?
    if (len(frame_list) == 0):
        raise IOError("No images read in. Check that filenames were indeed supplied.")
    elif (len(frame_list[0].shape) == 2 or len(frame_list[0].shape) == 3):
        width = frame_list[0].shape[0]
        height = frame_list[0].shape[1]
    else:  # pragma: nocover
        # some unexpected image shape
        raise IOError("Unexpected shape of image data. Found shape %s, but was expecting only 2 or 3 dimensions." % (frame_list[0].shape))

    # initialize videowriter object
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # type: ignore
    frame_size = (height, width)
    writer = cv2.VideoWriter(output_filename, fourcc, fps, frame_size)

    # write the frames
    if (progress_bar_disable is True):
        # no progress bar
        for i in range(0, len(frame_list)):
            writer.write(frame_list[i])
    else:
        # with progress bar
        for i in aurorax_obj._tqdm(range(0, len(frame_list)), desc="Encoding frames: ", unit="frames"):
            writer.write(frame_list[i])

    # close the writer
    writer.release()



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def __scale_data(data, min, max, top):
    # init
    bottom = 0

    # set top val
    #
    # NOTE: we only care about this if it's a uint array. If it's a double array, then we
    # check to make sure that a top was specified.
    if ("float" in str(data.dtype)):
        # this is float type, check that the top was specified
        if (top is None):
            raise ValueError("The top parameter must be specified when a float array is supplied")
    else:
        dtype_maxval = np.iinfo(data.dtype).max
        if (top is None):
            # derive values using dtype of data array
            top = dtype_maxval

        # check top
        if (top > dtype_maxval):
            raise ValueError("The top value must be less than or equal to %s" % (dtype_maxval))

    # set min and max
    if (min is None):
        cmin = data.min()
    else:
        cmin = float(min)
    if (max is None):
        cmax = data.max()
    else:
        cmax = float(max)

    # set scaling factor
    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError("The max value must be larger than the min value")
    elif cscale == 0:
        cscale = 1

    # do scaling
    scale = float(top - bottom) / cscale
    byte_data = (data - cmin) * scale + bottom
    scaled_data = (byte_data.clip(bottom, top) + 0.5).astype(data.dtype)

    # return
    return scaled_data


def scale_intensity(data, min, max, top, memory_saver):
    if (memory_saver is True):
        # Save original data shape
        input_shape = data.shape

        # determine if we are single or 3 channel
        if (len(data.shape) == 2):
            n_channels = 1
            data = data[:, :, np.newaxis]
        elif (len(data.shape) == 3):
            # single channel
            n_channels = 1
        elif (len(data.shape) == 4):
            # three channel
            n_channels = 3
        else:  # pragma: nocover
            raise ValueError("Unable to determine number of channels based on the supplied images. Make sure you are supplying a " +
                             "[rows,cols,images] or [rows,cols,channels,images] sized array.")

        # init destination array
        images_scaled = np.empty((data.shape), dtype=data.dtype)

        # cycle through each image
        for i in range(0, data.shape[-1]):
            if (n_channels == 1):
                images_scaled[:, :, i] = __scale_data(data[:, :, i], min, max, top)
            else:
                images_scaled[:, :, :, i] = __scale_data(data[:, :, :, i], min, max, top)

        # Reshape to confirm output is the same shape as input
        images_scaled = np.reshape(images_scaled, input_shape)

        # return
        return images_scaled
    else:
        # scale and return
        return __scale_data(data, min, max, top)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interact with the AuroraX search engine. This includes finding data sources, searching 
for conjunctions or ephemeris data, and uploading/managing your own data in the AuroraX platform.
"""

# pull in classes
from .location import Location
from .sources.classes.data_source import DataSource
from .availability.classes.availability_result import AvailabilityResult
from .ephemeris.classes.ephemeris import EphemerisData
from .ephemeris.classes.search import EphemerisSearch
from .data_products.classes.data_product import DataProductData
from .data_products.classes.search import DataProductSearch
from .conjunctions.classes.conjunction import Conjunction
from .conjunctions.classes.search import ConjunctionSearch
from .conjunctions.classes.criteria_block import (
    GroundCriteriaBlock,
    SpaceCriteriaBlock,
    EventsCriteriaBlock,
    CustomLocationsCriteriaBlock,
)
from .metadata_filters import MetadataFilter, MetadataFilterExpression

# pull in constants
from .sources.classes.data_source import (
    FORMAT_BASIC_INFO,
    FORMAT_BASIC_INFO_WITH_METADATA,
    FORMAT_FULL_RECORD,
    FORMAT_IDENTIFIER_ONLY,
    FORMAT_DEFAULT,
    SOURCE_TYPE_EVENT_LIST,
    SOURCE_TYPE_GROUND,
    SOURCE_TYPE_HEO,
    SOURCE_TYPE_LEO,
    SOURCE_TYPE_LUNAR,
)
from .data_products.classes.data_product import (
    DATA_PRODUCT_TYPE_KEOGRAM,
    DATA_PRODUCT_TYPE_MONTAGE,
    DATA_PRODUCT_TYPE_MOVIE,
    DATA_PRODUCT_TYPE_SUMMARY_PLOT,
    DATA_PRODUCT_TYPE_DATA_AVAILABILITY,
)
from .conjunctions.classes.conjunction import CONJUNCTION_TYPE_NBTRACE, CONJUNCTION_TYPE_SBTRACE

# imports for this file
from . import api as module_api
from .util import UtilManager
from .sources import SourcesManager
from .availability import AvailabilityManager
from .metadata import MetadataManager
from .requests import RequestsManager
from .ephemeris import EphemerisManager
from .data_products import DataProductsManager
from .conjunctions import ConjunctionsManager

__all__ = [
    "SearchManager",
    "FORMAT_BASIC_INFO",
    "FORMAT_BASIC_INFO_WITH_METADATA",
    "FORMAT_FULL_RECORD",
    "FORMAT_IDENTIFIER_ONLY",
    "FORMAT_DEFAULT",
    "SOURCE_TYPE_EVENT_LIST",
    "SOURCE_TYPE_GROUND",
    "SOURCE_TYPE_HEO",
    "SOURCE_TYPE_LEO",
    "SOURCE_TYPE_LUNAR",
    "DATA_PRODUCT_TYPE_KEOGRAM",
    "DATA_PRODUCT_TYPE_MONTAGE",
    "DATA_PRODUCT_TYPE_MOVIE",
    "DATA_PRODUCT_TYPE_SUMMARY_PLOT",
    "DATA_PRODUCT_TYPE_DATA_AVAILABILITY",
    "CONJUNCTION_TYPE_NBTRACE",
    "CONJUNCTION_TYPE_SBTRACE",
    "DataSource",
    "Location",
    "AvailabilityResult",
    "EphemerisData",
    "EphemerisSearch",
    "DataProductData",
    "DataProductSearch",
    "Conjunction",
    "ConjunctionSearch",
    "GroundCriteriaBlock",
    "SpaceCriteriaBlock",
    "EventsCriteriaBlock",
    "CustomLocationsCriteriaBlock",
    "MetadataFilter",
    "MetadataFilterExpression",
]


class SearchManager:
    """
    The SearchManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

        # initialize sub-modules
        self.__util = UtilManager()
        self.__api = module_api
        self.__sources = SourcesManager(self.__aurorax_obj)
        self.__availability = AvailabilityManager(self.__aurorax_obj)
        self.__metadata = MetadataManager(self.__aurorax_obj)
        self.__requests = RequestsManager(self.__aurorax_obj)
        self.__ephemeris = EphemerisManager(self.__aurorax_obj)
        self.__data_products = DataProductsManager(self.__aurorax_obj)
        self.__conjunctions = ConjunctionsManager(self.__aurorax_obj)

        # initialize class vars
        self.DataSource = DataSource
        self.Location = Location
        self.MetadataFilter = MetadataFilter
        self.MetadataFilterExpression = MetadataFilterExpression
        self.GroundCriteriaBlock = GroundCriteriaBlock
        self.SpaceCriteriaBlock = SpaceCriteriaBlock
        self.EventsCriteriaBlock = EventsCriteriaBlock
        self.CustomLocationsCriteriaBlock = CustomLocationsCriteriaBlock
        self.ConjunctionSearch = ConjunctionSearch
        self.EphemerisSearch = EphemerisSearch
        self.DataProductSearch = DataProductSearch
        self.EphemerisData = EphemerisData
        self.DataProductData = DataProductData

        # initialize static vars
        self.FORMAT_BASIC_INFO = FORMAT_BASIC_INFO
        self.FORMAT_BASIC_INFO_WITH_METADATA = FORMAT_BASIC_INFO_WITH_METADATA
        self.FORMAT_FULL_RECORD = FORMAT_FULL_RECORD
        self.FORMAT_IDENTIFIER_ONLY = FORMAT_IDENTIFIER_ONLY
        self.FORMAT_DEFAULT = FORMAT_DEFAULT
        self.SOURCE_TYPE_EVENT_LIST = SOURCE_TYPE_EVENT_LIST
        self.SOURCE_TYPE_GROUND = SOURCE_TYPE_GROUND
        self.SOURCE_TYPE_HEO = SOURCE_TYPE_HEO
        self.SOURCE_TYPE_LEO = SOURCE_TYPE_LEO
        self.SOURCE_TYPE_LUNAR = SOURCE_TYPE_LUNAR
        self.DATA_PRODUCT_TYPE_KEOGRAM = DATA_PRODUCT_TYPE_KEOGRAM
        self.DATA_PRODUCT_TYPE_MONTAGE = DATA_PRODUCT_TYPE_MONTAGE
        self.DATA_PRODUCT_TYPE_MOVIE = DATA_PRODUCT_TYPE_MOVIE
        self.DATA_PRODUCT_TYPE_SUMMARY_PLOT = DATA_PRODUCT_TYPE_SUMMARY_PLOT
        self.DATA_PRODUCT_TYPE_DATA_AVAILABILITY = DATA_PRODUCT_TYPE_DATA_AVAILABILITY
        self.CONJUNCTION_TYPE_NBTRACE = CONJUNCTION_TYPE_NBTRACE
        self.CONJUNCTION_TYPE_SBTRACE = CONJUNCTION_TYPE_SBTRACE

    # ------------------------------------------
    # properties for submodule managers
    # ------------------------------------------
    @property
    def util(self):
        """
        Access to the `util` submodule from within a PyAuroraX object.
        """
        return self.__util

    @property
    def api(self):
        """
        Access to the `api` submodule from within a PyAuroraX object.
        """
        return self.__api

    @property
    def sources(self):
        """
        Access to the `sources` submodule from within a PyAuroraX object.
        """
        return self.__sources

    @property
    def availability(self):
        """
        Access to the `availability` submodule from within a PyAuroraX object.
        """
        return self.__availability

    @property
    def metadata(self):
        """
        Access to the `metadata` submodule from within a PyAuroraX object.
        """
        return self.__metadata

    @property
    def requests(self):
        """
        Access to the `requests` submodule from within a PyAuroraX object.
        """
        return self.__requests

    @property
    def ephemeris(self):
        """
        Access to the `ephemeris` submodule from within a PyAuroraX object.
        """
        return self.__ephemeris

    @property
    def data_products(self):
        """
        Access to the `data_products` submodule from within a PyAuroraX object.
        """
        return self.__data_products

    @property
    def conjunctions(self):
        """
        Access to the `conjunctions` submodule from within a PyAuroraX object.
        """
        return self.__conjunctions



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Obtain contours in pixel coordinates from a skymap for plotting over CCD images.
"""

import datetime
import numpy as np
from typing import Optional, Union, Tuple
from ...data.ucalgary import Skymap
from ._azimuth import azimuth as func_azimuth
from ._elevation import elevation as func_elevation
from ._geo import geo as func_geo
from ._mag import mag as func_mag

__all__ = ["CCDContourManager"]


class CCDContourManager:
    """
    The CCDContourManager object is initialized within every PyAuroraX object. It acts as a way to 
    access the submodules and carry over configuration information in the super class.
    """

    def __init__(self):
        pass

    def azimuth(self,
                skymap: Skymap,
                constant_azimuth: Union[int, float],
                min_elevation: Optional[Union[int, float]] = None,
                max_elevation: Optional[Union[int, float]] = None,
                n_points: Optional[int] = None,
                remove_edge_cases: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Obtain CCD Coordinates of a line of constant latitude.

        Args:
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the CCD image data to generate contours for.

            constant_azimuth (int or float): 
                The azimuth angle, in degrees, to create contour of.

            min_elevation (int or float): 
                Optionally specify the elevation angle at which contour begins. Defaults to 5.
            
            max_elevation (int or float): 
                Optionally specify the elevation angle at which contour begins. Defaults to 90.

            n_points (int or float): 
                Optionally specify the number of points used to define a contour. By default
                a reasonable value is selected automatically.

            remove_edge_cases (bool): 
                Due to the nature of skymaps, often, around the edge of CCD data, contours will
                have often undesired behaviour due to being bounded within the CCD range. The result
                is flattened contours along the edge of CCD boundaries. This is completely expected,
                and these points are removed by default, completely for aesthetic choices. Set this 
                keyword to False to keep all points in the contour.
                
        Returns:
            A tuple (x_pix, y_pix) of numpy arrays containing the coordinates, in pixel units, of
            the azimuth contour.

        Raises:
            ValueError: invalid azimuth supplied.
        """
        return func_azimuth(skymap, constant_azimuth, min_elevation, max_elevation, n_points, remove_edge_cases)

    def elevation(self,
                  skymap: Skymap,
                  constant_elevation: Union[int, float],
                  n_points: Optional[int] = None,
                  remove_edge_cases: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Obtain CCD Coordinates of a line of constant elevation.

        Args:
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the CCD image data to generate contours for.

            constant_elevation (int or float): 
                The elevation angle, in degrees from the horizon, to create contour of.

            n_points (int or float): 
                Optionally specify the number of points used to define a contour. By default
                a reasonable value is selected automatically.

            remove_edge_cases (bool): 
                Due to the nature of skymaps, often, around the edge of CCD data, contours will
                have often undesired behaviour due to being bounded within the CCD range. The result
                is flattened contours along the edge of CCD boundaries. This is completely expected,
                and these points are removed by default, completely for aesthetic choices. Set this 
                keyword to False to keep all points in the contour.
                
        Returns:
            A tuple (x_pix, y_pix) of numpy arrays containing the coordinates, in pixel units, of
            the elevation contour.

        Raises:
            ValueError: invalid elevation supplied.
        """
        return func_elevation(skymap, constant_elevation, n_points, remove_edge_cases)

    def geo(self,
            skymap: Skymap,
            altitude_km: Union[int, float],
            contour_lats: Optional[Union[np.ndarray, list]] = None,
            contour_lons: Optional[Union[np.ndarray, list]] = None,
            constant_lat: Optional[Union[float, int]] = None,
            constant_lon: Optional[Union[float, int]] = None,
            n_points: Optional[int] = None,
            remove_edge_cases: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Obtain CCD Coordinates of a line of constant geographic latitude, constant geographic longitude, or a custom contour
        defined in geographic coordinates.

        Args:
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the CCD image data to generate contours for.

            altitude_km (int or float): 
                The altitude of the image data to create contours for, in kilometers.

            contour_lats (ndarray or list): 
                    Sequence of geographic latitudes defining a contour.
                
            contour_lons (ndarray or list): 
                Sequence of geographic longitudes defining a contour.

            constant_lat (float or int): 
                Geographic Latitude at which to create line of constant latitude.
            
            constant_lon (float or int): 
                Geographic Longitude at which to create line of constant longitude.

            n_points (int or float): 
                Optionally specify the number of points used to define a contour. By default
                a reasonable value is selected automatically.
            
            remove_edge_cases (bool): 
                Due to the nature of skymaps, often, around the edge of CCD data, contours will
                have often undesired behaviour due to being bounded within the CCD range. The result
                is flattened contours along the edge of CCD boundaries. This is completely expected,
                and these points are removed by default, completely for aesthetic choices. Set this 
                keyword to False to keep all points in the contour.
                
        Returns:
            A tuple (x_pix, y_pix) of numpy arrays containing the coordinates, in pixel units, of
            the elevation contour.

        Raises:
            ValueError: invalid elevation supplied.
        """
        return func_geo(skymap, altitude_km, contour_lats, contour_lons, constant_lat, constant_lon, n_points, remove_edge_cases)

    def mag(self,
            skymap: Skymap,
            timestamp: datetime.datetime,
            altitude_km: Union[int, float],
            contour_lats: Optional[Union[np.ndarray, list]] = None,
            contour_lons: Optional[Union[np.ndarray, list]] = None,
            constant_lat: Optional[Union[float, int]] = None,
            constant_lon: Optional[Union[float, int]] = None,
            n_points: Optional[int] = None,
            remove_edge_cases: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Obtain CCD Coordinates of a line of constant magnetic latitude, constant magnetic longitude, or a custom contour
        defined in magnetic coordinates.

        Args:
            skymap (pyaurorax.data.ucalgary.Skymap): 
                The skymap corresponding to the CCD image data to generate contours for.

            timestamp (datetime.datetime): 
                The timestamp used for AACGM Conversions.

            altitude_km (int or float): 
                The altitude of the image data to create contours for, in kilometers.

            contour_lats (ndarray or list): 
                    Sequence of magnetic latitudes defining a contour.
                
            contour_lons (ndarray or list): 
                Sequence of magnetic longitudes defining a contour.

            constant_lat (float or int): 
                Magnetic Latitude at which to create line of constant latitude.
            
            constant_lon (float or int): 
                Magnetic Longitude at which to create line of constant longitude.

            n_points (int or float): 
                Optionally specify the number of points used to define a contour. By default
                a reasonable value is selected automatically.
            
            remove_edge_cases (bool): 
                Due to the nature of skymaps, often, around the edge of CCD data, contours will
                have often undesired behaviour due to being bounded within the CCD range. The result
                is flattened contours along the edge of CCD boundaries. This is completely expected,
                and these points are removed by default, completely for aesthetic choices. Set this 
                keyword to False to keep all points in the contour.
                
        Returns:
            A tuple (x_pix, y_pix) of numpy arrays containing the coordinates, in pixel units, of
            the elevation contour.

        Raises:
            ValueError: invalid elevation supplied.
        """
        return func_mag(skymap, timestamp, altitude_km, contour_lats, contour_lons, constant_lat, constant_lon, n_points, remove_edge_cases)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from .availability import commands as availability_commands
from .conjunctions import commands as conjunctions_commands
from .data_products import commands as data_products_commands
from .ephemeris import commands as ephemeris_commands
from .sources import commands as sources_commands
from .util import commands as util_commands


@click.group("search", help="Search engine commands")
def search_group():
    pass


search_group.add_command(availability_commands.availability_group)
search_group.add_command(conjunctions_commands.conjunctions_group)
search_group.add_command(data_products_commands.data_products_group)
search_group.add_command(ephemeris_commands.ephemeris_group)
search_group.add_command(sources_commands.sources_group)
search_group.add_command(util_commands.utility_group)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import matplotlib.pyplot as plt
from .._util import show_warning


def display(image, cmap, figsize, aspect, colorbar, title, returnfig, savefig, savefig_filename, savefig_quality):
    # check return mode
    if (returnfig is True and savefig is True):
        raise ValueError("Only one of returnfig or savefig can be set to True")
    if (returnfig is True and (savefig_filename is not None or savefig_quality is not None)):
        show_warning("The figure will be returned, but a savefig option parameter was supplied. Consider " +
                     "removing the savefig option parameter(s) as they will be ignored.",
                     stacklevel=1)
    elif (savefig is False and (savefig_filename is not None or savefig_quality is not None)):
        show_warning("A savefig option parameter was supplied, but the savefig parameter is False. The " +
                     "savefig option parameters will be ignored.",
                     stacklevel=1)

    # init figure
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes((0, 0, 1, 1))
    ax.set_axis_off()

    # plot data
    im = ax.imshow(image, cmap=cmap, origin="lower", aspect=aspect)

    # show colour bar
    if (colorbar is True):
        fig.colorbar(im, ax=ax)

    # show title
    if (title is not None):
        ax.set_title(title)

    # save figure or show it
    if (savefig is True):
        # check that filename has been set
        if (savefig_filename is None):
            raise ValueError("The savefig_filename parameter is missing, but required since savefig was set to True.")

        # save the figure
        f_extension = os.path.splitext(savefig_filename)[-1].lower()
        if (".jpg" == f_extension or ".jpeg" == f_extension):
            # check quality setting
            if (savefig_quality is not None):  # pragma: nocover-ok
                plt.savefig(savefig_filename, quality=savefig_quality, bbox_inches="tight")
            else:
                plt.savefig(savefig_filename, bbox_inches="tight")
        else:
            if (savefig_quality is not None):
                # quality specified, but output filename is not a JPG, so show a warning
                show_warning("The savefig_quality parameter was specified, but is only used for saving JPG files. The " +
                             "savefig_filename parameter was determined to not be a JPG file, so the quality will be ignored",
                             stacklevel=1)
            plt.savefig(savefig_filename, bbox_inches="tight")

        # clean up by closing the figure
        plt.close(fig)
    elif (returnfig is True):
        # return the figure and axis objects
        return (fig, ax)
    else:
        # show the figure
        plt.show(fig)

        # cleanup by closing the figure
        plt.close(fig)

    # return
    return None
